# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import glob
import logging
import os
import time
from typing import Dict, List, Optional, Sequence, Tuple

import numpy as np
from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset import Dataset
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.environment import Environment

from geti_sdk.data_models import TaskType
from geti_sdk.utils import get_dict_key_from_value


class DatumaroDataset(object):
    """
    Wrapper for interacting with the datumaro dataset, contains some example
    functions for dataset operations that can be carried out prior to importing the
    dataset into an Intel® Geti™ project.
    """

    def __init__(self, dataset_format: str, dataset_path: str):
        """
        Initialize the datumaro dataset.

        :param dataset_format: string containing the format of the dataset, i.e.
            'voc', 'coco', 'imagenet', etc.
        :param dataset_path: Path to the dataset on local disk
        """
        self.dataset_format = dataset_format
        self.dataset_path = dataset_path
        self.dataset, self.environment = self.create_datumaro_dataset()
        self._subset_names = self.dataset.subsets().keys()
        self._filtered_categories = self.dataset.categories()[AnnotationType.label]

    def prepare_dataset(
        self, task_type: TaskType, previous_task_type: Optional[TaskType] = None
    ) -> Dataset:
        """
        Prepare the dataset for uploading to Intel Geti.

        :param task_type: TaskType to prepare the dataset for
        :param previous_task_type: Optional type of the (trainable) task preceding
            the current task in the pipeline. This is only used for global tasks
        """
        if task_type.is_detection and task_type != TaskType.ROTATED_DETECTION:
            new_dataset = self.dataset.transform(
                self.dataset.env.transforms.get("shapes_to_boxes")
            )
            logging.info("Annotations have been converted to boxes")
        elif task_type.is_segmentation or task_type == TaskType.ROTATED_DETECTION:
            converted_dataset = self.dataset.transform(
                self.dataset.env.transforms.get("masks_to_polygons")
            )
            new_dataset = converted_dataset.filter(
                '/item/annotation[type="polygon"]', filter_annotations=True
            )
            logging.info("Annotations have been converted to polygons")
        elif task_type.is_global:
            if previous_task_type is not None and (
                previous_task_type.is_segmentation
                or previous_task_type == TaskType.ROTATED_DETECTION
            ):
                converted_dataset = self.dataset.transform(
                    self.dataset.env.transforms.get("masks_to_polygons")
                )
                new_dataset = converted_dataset.filter(
                    '/item/annotation[type="polygon"]', filter_annotations=True
                )
            else:
                new_dataset = self.dataset.transform(
                    self.dataset.env.transforms.get("shapes_to_boxes")
                )
            logging.info(f"{str(task_type).capitalize()} dataset prepared.")
        else:
            raise ValueError(f"Unsupported task type {task_type}")
        return new_dataset

    def set_dataset(self, dataset: Dataset):
        """
        Set the dataset for this DatumaroDataset instance.

        :param dataset:
        :return:
        """
        self.dataset = dataset
        self.environment = dataset.env
        self._subset_names = self.dataset.subsets().keys()

    @property
    def categories(self) -> LabelCategories:
        """
        Return the LabelCategories in the dataset.
        """
        return self._filtered_categories

    @property
    def label_names(self) -> List[str]:
        """
        Return a list of all label names in the dataset.
        """
        return [item.name for item in self.categories]

    @property
    def label_mapping(self) -> Dict[int, str]:
        """
        Return the mapping of label index to label name.
        """
        return {value: key for key, value in self.categories._indices.items()}

    @property
    def image_names(self) -> List[str]:
        """
        Return the list of media names included in the dataset.
        """
        return [item.id for item in self.dataset]

    def create_datumaro_dataset(self) -> Tuple[Dataset, Environment]:
        """
        Initialize a datumaro dataset.
        """
        t_start = time.time()
        dataset = Dataset.import_from(
            path=self.dataset_path, format=self.dataset_format
        )
        logging.info(
            f"Datumaro dataset consisting of {len(dataset)} items in "
            f"{self.dataset_format} format was loaded from {self.dataset_path}"
        )
        logging.info(
            f"Datumaro dataset was created in {time.time() - t_start:.1f} seconds"
        )
        return dataset, dataset.env

    def remove_unannotated_items(self):
        """
        Keep only annotated images.
        """
        self.dataset = self.dataset.select(lambda item: len(item.annotations) != 0)

    def filter_items_by_labels(self, labels: Sequence[str], criterion="OR") -> None:
        """
        Retain only those items with annotations in the list of labels passed.

        :param labels: List of labels to filter on
        :param criterion: Filter criterion, currently "OR", "NOT", "AND" and "XOR" are
            implemented
        """
        label_map = self.label_mapping
        # Sanity check for filtering
        for label in labels:
            if label not in list(label_map.values()):
                raise ValueError(
                    f"Cannot filter on label {label} because this is not in the "
                    f"dataset."
                )

        if labels:

            def select_function(dataset_item: DatasetItem, labels: List[str]):
                # Filter function to apply to each item in the dataset
                item_labels = [label_map[x.label] for x in dataset_item.annotations]
                matches = []
                for label in labels:
                    if label in item_labels:
                        if criterion == "OR":
                            return True
                        elif criterion in ["AND", "NOT", "XOR"]:
                            matches.append(True)
                        else:
                            raise ValueError(
                                'Invalid filter criterion, please select "OR", "NOT", '
                                '"XOR", or "AND".'
                            )
                    else:
                        matches.append(False)
                if criterion == "AND":
                    return all(matches)
                elif criterion == "NOT":
                    return not any(matches)
                elif criterion == "XOR":
                    return np.sum(matches) == 1

            # Messy way to manually keep track of labels and indices, must be a
            # better way in Datumaro but haven't found it yet
            label_categories = LabelCategories.from_iterable(labels)
            new_labelmap = {}
            for label in labels:
                label_key = get_dict_key_from_value(label_map, label)
                new_labelmap[label_key] = label
            label_categories._indices = {v: k for k, v in new_labelmap.items()}
            new_categories = label_categories
            # Filter and create a new dataset to update the dataset categories
            self.dataset = Dataset.from_iterable(
                self.dataset.select(lambda item: select_function(item, labels)),
                categories=self.dataset.categories(),
                env=self.dataset.env,
            )
            logging.info(
                f"After filtering, dataset with labels {labels} contains "
                f"{len(self.dataset)} items."
            )
            self._filtered_categories = new_categories

    def __get_item_by_id_from_subsets(
        self, datum_id: str, search_by_name: bool = False
    ) -> Optional[DatasetItem]:
        """
        Search all subsets for the item with id `datum_id`

        :param datum_id: Datumaro id of the item to retrieve
        :param search_by_name: True to search for the image by filename as well, in
            addition to searching within the datumaro dataset
        :return: Dataset item with the given id, or None if the item was not found
        """
        ds_item: Optional[DatasetItem] = None
        for subset_name in self._subset_names:
            ds_item = self.dataset.get(id=datum_id, subset=subset_name)
        if search_by_name and ds_item is None:
            for subset_name in self._subset_names:
                image_relative_path = self.__search_image_by_filename(
                    image_filename=datum_id, subset_name=subset_name
                )
                if image_relative_path is not None:
                    ds_item = self.dataset.get(
                        id=image_relative_path, subset=subset_name
                    )
        return ds_item

    def __search_image_by_filename(
        self, image_filename: str, subset_name: str
    ) -> Optional[str]:
        """
        Search for an image with name `image_filename` inside the directory tree in
        the data path.

        :param image_filename: Filename of the image to search for (without extension!)
        :param subset_name: Name of the subset which the image is in
        :return: Datumaro id (expressed as a relative unix-style path) for the image,
            defined with respect to the subset it is in. If no matches are found for the
            filename, this method returns None
        """
        matches = glob.glob(
            os.path.join(
                self.dataset_path,
                "**",
                "images",
                subset_name,
                "**",
                f"{image_filename}.*",
            ),
            recursive=True,
        )
        if len(matches) > 1:
            logging.warning(
                f"Multiple images with filename '{image_filename}' found in dataset "
                f"subset '{subset_name}', unable to uniquely identify dataset item"
            )
            return None
        elif len(matches) == 0:
            return None
        relative_path = matches[0].split(f"{subset_name}{os.path.sep}")[-1]
        path = os.path.normpath(relative_path)
        path = os.path.splitext(path)[0]
        path_components = path.split(os.path.sep)
        return "/".join(path_components)

    def get_item_by_id(self, datum_id: str) -> DatasetItem:
        """
        Return the dataset item by its id.

        :param datum_id: Datumaro id of the item to retrieve
        :raises: ValueError if no item by that id was found.
        :return: DatasetItem with the given id
        """
        ds_item = self.__get_item_by_id_from_subsets(
            datum_id=datum_id, search_by_name=True
        )
        if ds_item is None:
            raise ValueError(
                f"Unable to identify dataset item with id {datum_id} in the dataset. "
                f"Please try to simplify the internal filestructure of the dataset, "
                f"and make sure that images names (within subsets) are unique."
            )
        return ds_item



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os
import zipfile
from enum import Enum
from typing import Optional, Tuple

from geti_sdk.demos.constants import DEFAULT_DATA_PATH

from .download_helpers import download_file, validate_hash

DEFAULT_COCO_PATH = os.path.join(DEFAULT_DATA_PATH, "coco")


class COCOSubset(Enum):
    """
    Enum representing a certain subset of the MS COCO dataset.
    """

    TRAIN2014 = "train2014"
    VAL2014 = "val2014"
    TEST2014 = "test2014"
    TEST2015 = "test2015"
    TRAIN2017 = "train2017"
    VAL2017 = "val2017"
    TEST2017 = "test2017"
    UNLABELED2017 = "unlabeled2017"

    def __str__(self):
        """
        Return string representation of the COCOSubset instance.
        """
        return self.value

    def get_annotations(self) -> Optional[str]:
        """
        Return the name of the annotation folder for the subset, if any.

        :return: String containing the name of the annotation folder for the subset.
            If subset does not have annotations, returns None.
        """
        if self == COCOSubset.VAL2017 or self == COCOSubset.TRAIN2017:
            return "trainval2017"
        elif self == COCOSubset.VAL2014 or self == COCOSubset.TRAIN2014:
            return "trainval2014"
        elif (
            self == COCOSubset.TEST2017
            or self == COCOSubset.TEST2015
            or self == COCOSubset.TEST2014
            or self == COCOSubset.UNLABELED2017
        ):
            return None

    def get_hashes(self) -> Tuple[Optional[str], Optional[str]]:
        """
        Return the expected sha256 hashes for the .zip files containing the images
        and the annotations of the dataset

        Note: If an expected hash is not known for a subset, it will be returned as
            `None`

        :return: Tuple holding the hash digests for the archives, structured like:
            (image_zip, annotation_zip).
        """
        if self == COCOSubset.VAL2017:
            return (
                "9ea554bcf9e6f88876b1157ab38247eb7c1c57564c05c7345a06ac479c6e7a3b9c3825150c189d7d3f2e807c95fd0e07fe90161c563591038e697c846ac76007",
                "3f00c90323ee745b37a9ac040d00f170d49695ed9ffc1d8e0fbd4c5e2d8e9c697fd822b2022df552da5f1892dbcaeb68788416a347b05a20035ed0686f0e1f66",
            )
        else:
            return None, None


def directory_has_coco_subset(target_folder: str, coco_subset: COCOSubset) -> bool:
    """
    Check if a certain directory contains a particular subset of the coco dataset.

    :param target_folder: Directory to check
    :param coco_subset: Subset to look for
    :return: True if the directory contains the subset, False otherwise
    """
    if not os.path.exists(target_folder):
        return False
    required_dirs = ["images", "annotations"]
    for directory in required_dirs:
        if directory not in os.listdir(target_folder):
            return False
    image_dir = os.path.join(target_folder, "images")
    subset_image_dir = os.path.join(image_dir, str(coco_subset))
    if not os.path.exists(subset_image_dir):
        return False
    if len(os.listdir(subset_image_dir)) == 0:
        return False
    annotations = coco_subset.get_annotations()
    if annotations is not None:
        annotations_dir_content = os.listdir(os.path.join(target_folder, "annotations"))
        if len(annotations_dir_content) == 0:
            return False
        for filename in annotations_dir_content:
            if f"instances_{str(coco_subset)}" in filename:
                return True
        return False
    return True


def get_coco_dataset_from_path(
    target_folder: str = "data",
    coco_subset: Optional[COCOSubset] = None,
    verbose: bool = False,
) -> str:
    """
    Check if a coco dataset exists in the directory at `target_folder`.
    If no such directory exists, this method will attempt to download the COCO
    dataset to the designated folder.

    :param target_folder: Folder to get the COCO dataset from, or to download the
        dataset into
    :param coco_subset: Optional Subset to download. If passed as None, the script
        will check if any of the coco subsets exists in the folder, and will not
        download any data if at least one subset is found.
    :param verbose: True to print verbose output, False for silent mode
    :return: Path to the COCO dataset
    """
    os.makedirs(target_folder, exist_ok=True, mode=0o770)
    found_subset = None

    if coco_subset is None:
        for subset in COCOSubset:
            if subset.get_annotations() is not None and directory_has_coco_subset(
                target_folder, subset
            ):
                found_subset = subset
                break
        if found_subset is None:
            found_subset = COCOSubset.VAL2017
    else:
        found_subset = coco_subset

    if directory_has_coco_subset(target_folder=target_folder, coco_subset=found_subset):
        if verbose:
            logging.info(
                f"COCO dataset (subset: {str(found_subset)}) found at path "
                f"{target_folder}"
            )
        return target_folder
    else:
        logging.info(
            f"COCO dataset was not found at path {target_folder}, making an "
            f"attempt to download the data."
        )

    image_url = f"http://images.cocodataset.org/zips/{str(found_subset)}.zip"
    annotations_name = found_subset.get_annotations()
    if annotations_name is not None:
        annotations_url = (
            f"http://images.cocodataset.org/annotations/"
            f"annotations_{annotations_name}.zip"
        )
    else:
        if verbose:
            logging.info(
                f"Unable to download annotations for COCO subset {found_subset}. "
                f"Downloading images only"
            )
        annotations_url = None

    # Download the zip files
    image_zip = download_file(image_url, target_folder, check_valid_archive=True)
    if annotations_url is not None:
        annotations_zip = download_file(
            annotations_url, target_folder, check_valid_archive=True
        )
    else:
        annotations_zip = None

    # Create directories for images and annotations
    image_dir = os.path.join(target_folder, "images")
    os.makedirs(image_dir, exist_ok=True, mode=0o770)
    hashes = found_subset.get_hashes()
    zip_to_extraction_mapping = {
        image_zip: {"directory": image_dir, "expected_hash": hashes[0]}
    }

    if annotations_zip is not None:
        annotations_dir = os.path.join(target_folder, "annotations")
        os.makedirs(annotations_dir, exist_ok=True, mode=0o770)
        zip_to_extraction_mapping.update(
            {annotations_zip: {"directory": target_folder, "expected_hash": hashes[1]}}
        )
    else:
        annotations_dir = None

    # Extract images and annotations
    for zipfile_path, data_dictionary in zip_to_extraction_mapping.items():
        if verbose:
            logging.info(f"Extracting {zipfile_path}...")
        if data_dictionary["expected_hash"] is not None:
            validate_hash(zipfile_path, data_dictionary["expected_hash"])
        with zipfile.ZipFile(zipfile_path, "r") as zip_ref:
            zip_ref.extractall(data_dictionary["directory"])
        os.remove(zipfile_path)

    image_subset_names = os.listdir(image_dir)
    if annotations_dir is not None:
        # Remove unused annotations
        for filename in os.listdir(annotations_dir):
            filepath = os.path.join(annotations_dir, filename)
            if "instances" not in filename:
                os.remove(filepath)
                continue
            annotation_subset = os.path.splitext(filename.split("_")[1])[0]
            if annotation_subset not in image_subset_names:
                os.remove(filepath)

    if verbose:
        logging.info("COCO dataset downloaded and extracted successfully.")
    return target_folder


def get_coco_dataset(dataset_path: Optional[str] = None, verbose: bool = False) -> str:
    """
    Check if the COCO dataset is present at the specified path. If not,
    this method will attempt to download the dataset to the path specified.

    If no path is passed, this method will check or create the default path: the
    folder 'data' in the top level of the geti-sdk package.

    :param dataset_path: Path to check against.
    :param verbose: True to print detailed output, False to check silently
    :return: Path to the COCO dataset
    """
    if dataset_path is None:
        dataset_path = DEFAULT_COCO_PATH
    return get_coco_dataset_from_path(dataset_path, verbose=verbose)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `http_session` package acts as a wrapper around `requests.Session` and handles all
http requests to the GETi cluster.

Normally it is not necessary to interact with the classes in this package directly, as
they are instantiated through the main :py:class:`geti_sdk.Geti` class.

However, a GetiSession can be established directly using the following code snippet:

.. code-block:: python

   from geti_sdk.http_session import ServerCredentialConfig, GetiSession

   config = ServerCredentialConfig(
     host="https://0.0.0.0", username="dummy_user", password="dummy_password"
   )
   session = GetiSession(server_config=config)

Initializing the session will perform authentication on the GETi server.

Module contents
---------------

.. autoclass:: geti_sdk.http_session.geti_session.GetiSession
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: geti_sdk.http_session.server_config.ServerTokenConfig
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: geti_sdk.http_session.server_config.ServerCredentialConfig
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: geti_sdk.http_session.exception.GetiRequestException
   :members:
   :undoc-members:
   :show-inheritance:

"""

from .exception import GetiRequestException
from .geti_session import GetiSession
from .server_config import ServerCredentialConfig, ServerTokenConfig

__all__ = [
    "GetiSession",
    "ServerTokenConfig",
    "ServerCredentialConfig",
    "GetiRequestException",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict, Optional, TypeVar

KeyType = TypeVar("KeyType")


def get_dict_key_from_value(input_dict: Dict[KeyType, Any], value) -> Optional[KeyType]:
    """
    Return the key associated with `value` in a dictionary `input_dict`. If the value
    is not found in the dictionary, returns None.

    :param input_dict: Dictionary to search in
    :param value: value to search for
    :return: key associated with value if value is in the input_dict, None otherwise
    """
    keys = list(input_dict.keys())
    values = list(input_dict.values())
    try:
        index = values.index(value)
    except ValueError:
        return None
    return keys[index]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from geti_sdk.http_session import GetiSession


def get_default_workspace_id(rest_session: GetiSession) -> str:
    """
    Return the id of the default workspace on the cluster.

    :param rest_session: HTTP session to the cluser
    :return: string containing the id of the default workspace
    """
    workspaces = rest_session.get_rest_response(url="workspaces", method="GET")
    if isinstance(workspaces, list):
        workspace_list = workspaces
    elif isinstance(workspaces, dict):
        workspace_list = workspaces["workspaces"]
    else:
        raise ValueError(
            f"Unexpected response from cluster: {workspaces}. Expected to receive a "
            f"dictionary containing workspace data."
        )
    default_workspace_names = ["default", "default workspace", "default_workspace"]
    default_workspace = next(
        (
            workspace
            for workspace in workspace_list
            if workspace["name"].lower() in default_workspace_names
        )
    )
    return default_workspace["id"]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `prediction_visualization` package provides classes for visualizing models predictions and media annotations.
Aditionally, shortend interface to this package is available through the :py:mod:`~geti_sdk.utils.plot_helpers` module.

The main :py:class:`~geti_sdk.prediction_visualization.visualizer.Visualizer` class is a flexible utility class for working
with Geti-SDK Prediction and Annotation object. You can initialize the Visualizer with the desired settings and then use it to draw
the annotations on the input image.

.. code-block:: python

   from geti_sdk import Visualizer

   visualizer = Visualizer(
      show_labels=True,
      show_confidence=True,
      show_count=False,
   )

   # Obtain a prediction from the Intel Geti platfor server or a local deployment.
   ...

   # Visualize the prediction on the input image.
   result = visualizer.draw(
      numpy_image,
      prediction,
      fill_shapes=True,
      confidence_threshold=0.4,
   )
   visualizer.show_in_notebook(result)

In case the Prediction was generated with a model that supports explainable AI functionality, the Visualizer can also display
the explanation for the prediction.

.. code-block:: python
   image_with_saliency_map = visualizer.explain_label(
      numpy_image,
      prediction,
      label_name="Cat",
      opacity=0.5,
      show_predictions=True,
   )
   visualizer.save_image(image_with_saliency_map, "./explained_prediction.jpg")
   visualizer.show_window(image_with_saliency_map)  # When called in a script

Module contents
---------------

.. automodule:: geti_sdk.prediction_visualization.visualizer
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: geti_sdk.prediction_visualization.shape_drawer.ShapeDrawer
   :members:
   :undoc-members:
   :show-inheritance:
"""

from .visualizer import Visualizer

__all__ = ["Visualizer"]



import time
from typing import Optional, Union

from geti_sdk.data_models import Job, Task
from geti_sdk.data_models.algorithms import Algorithm
from geti_sdk.http_session import GetiRequestException
from geti_sdk.rest_clients import TrainingClient

from .enums import SdkTestMode


def attempt_to_train_task(
    training_client: TrainingClient,
    task: Union[int, Task],
    test_mode: SdkTestMode = SdkTestMode.OFFLINE,
    algorithm: Optional[Algorithm] = None,
) -> Job:
    """
    Attempts to train the `task` (either a task or the index of the task in the list
    of trainable tasks for the project), by issuing a 'train_task' command from the
    training_client.

    If the initial train request fails with a 'project_not_train_ready' error code,
    this method will attempt to issue the command again (max 5 attempts) after
    a short sleep period.

    :param training_client: TrainingClient to use to train the task
    :param task: Task or index of task to be trained
    :param test_mode: SdkTestMode in which the training is attempted. If set to
        SdkTestMode.OFFLINE, no sleep period is used between attempts
    :return: Training job that was created by the training request
    """
    job: Optional[Job] = None
    not_ready_response = "project_not_train_ready"
    n_attempts = 10

    for i in range(n_attempts):
        try:
            job = training_client.train_task(task, algorithm=algorithm)
            break
        except GetiRequestException as error:
            if error.response_error_code != not_ready_response:
                raise error
        if test_mode != SdkTestMode.OFFLINE:
            time.sleep(5)

    if job is not None:
        return job
    else:
        raise ValueError(
            f"Unable to start training on server, received '{not_ready_response}' on "
            f"each request for {n_attempts} calls in a row. Please make sure that the "
            f"project is ready to start training the task."
        )


def await_training_start(
    test_mode: SdkTestMode, training_client: TrainingClient, timeout: int = 300
) -> None:
    """
    Wait for the project to start training. This method will not trigger any training
    jobs, but merely await the project addressed by the `training_client` to start
    training.

    This method will poll the server in 10 second intervals, until training is started

    :param test_mode: SdkTestMode in which the training is attempted. If set to
        SdkTestMode.OFFLINE, this method does not do any waiting
    :param training_client: Training client for the project that is expected to start
        training
    :param timeout: Maximum amount of time (in seconds) to wait for training to start
    """
    if test_mode != SdkTestMode.OFFLINE:
        t_start = time.time()
        is_training = training_client.is_training()
        while not is_training and time.time() - t_start < timeout:
            is_training = training_client.is_training()
            time.sleep(10)
    return None



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Optional, Sequence, Type, TypeVar, Union

from requests import Response
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk.annotation_readers import AnnotationReader
from geti_sdk.data_models import (
    AnnotationKind,
    AnnotationScene,
    Dataset,
    Image,
    Project,
    Video,
    VideoFrame,
)
from geti_sdk.data_models.containers.media_list import MediaList
from geti_sdk.data_models.label import Label
from geti_sdk.data_models.media import MediaInformation, MediaItem
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_clients.dataset_client import DatasetClient
from geti_sdk.rest_converters import AnnotationRESTConverter

AnnotationReaderType = TypeVar("AnnotationReaderType", bound=AnnotationReader)
MediaType = TypeVar("MediaType", Image, Video)


class BaseAnnotationClient:
    """
    Class to up- or download annotations for 2d media to an existing project.
    """

    def __init__(
        self,
        session: GetiSession,
        workspace_id: str,
        project: Project,
        annotation_reader: Optional[AnnotationReaderType] = None,
    ):
        self.session = session
        self.workspace_id = workspace_id
        self.annotation_reader = annotation_reader
        self._project = project
        if annotation_reader is None:
            label_mapping = None
        else:
            label_mapping = self.__get_label_mapping(project)
        self._label_mapping = label_mapping
        self._dataset_client = DatasetClient(
            session=session, project=project, workspace_id=workspace_id
        )

    def _get_all_media_by_type(
        self, media_type: Type[MediaType]
    ) -> MediaList[MediaType]:
        """
        Get a list holding all media entities of type `media_type` in the project.

        :param media_type: Type of media item to retrieve. Can be 'Image' or 'Video'
        :return: MediaList holding all media of a certain type in the project
        """
        datasets = self._dataset_client.get_all_datasets()
        media_list = MediaList[media_type]([])
        for dataset in datasets:
            media_list.extend(
                self._get_all_media_in_dataset_by_type(
                    media_type=media_type, dataset=dataset
                )
            )
        return media_list

    def _get_all_media_in_dataset_by_type(
        self, media_type: Type[MediaType], dataset: Dataset
    ) -> MediaList[MediaType]:
        """
        Return a list of all media items of type `media_type` in the dataset `dataset`

        :param media_type: Type of media item to retrieve. Can be 'Image' or 'Video'
        :param dataset: Dataset to retrieve media items for
        :return: List of all media items of a certain media_type in the dataset
        """
        if media_type == Image:
            media_name = "images"
            single_media_name = "image"
        elif media_type == Video:
            media_name = "videos"
            single_media_name = "video"
        else:
            raise ValueError(f"Invalid media type specified: {media_type}.")

        url = (
            f"workspaces/{self.workspace_id}/projects/{self._project.id}"
            f"/datasets/{dataset.id}/media:query?limit=100"
        )
        data = {
            "condition": "and",
            "rules": [
                {
                    "field": "MEDIA_TYPE",
                    "operator": "EQUAL",
                    "value": single_media_name,
                }
            ],
        }
        response = self.session.get_rest_response(url=url, method="POST", data=data)
        total_number_of_media: int = response[f"total_matched_{media_name}"]

        raw_media_list: List[Dict[str, Any]] = []
        while len(raw_media_list) < total_number_of_media:
            for media_item_dict in response["media"]:
                raw_media_list.append(media_item_dict)
            if "next_page" in response.keys():
                response = self.session.get_rest_response(
                    url=response["next_page"],
                    method="POST",
                    data=data,
                    include_organization_id=False,
                )
        return MediaList.from_rest_list(
            rest_input=raw_media_list, media_type=media_type
        )

    def __get_label_mapping(self, project: Project) -> Dict[str, str]:
        """
        Get the mapping of the label names to the label ids for the project

        :param project:
        :return: Dictionary containing the label names as keys and the label ids as
            values
        """
        project_label_name_to_label: Dict[str, Label] = {}
        for label in project.pipeline.get_all_labels():
            if label.is_empty:
                # We perform a casefold on the label name to ensure that we can match
                # the empy labels from projects created in older versions of the
                # Intel Geti platform.
                label_name = label.name.casefold()
            else:
                label_name = label.name
            project_label_name_to_label[label_name] = label

        source_label_names = set(self.annotation_reader.get_all_label_names())
        source_label_name_to_project_label_id: Dict[str, str] = {}
        # We include the project label names in the mapping, as we want to ensure that
        # we can match the labels from the source to the project labels.
        for source_label_name in source_label_names.union(
            project_label_name_to_label.keys()
        ):
            if source_label_name in project_label_name_to_label:
                label_id = project_label_name_to_label[source_label_name].id
            elif (
                source_label_name.casefold() in project_label_name_to_label
                and project_label_name_to_label[source_label_name.casefold()].is_empty
            ):
                label_id = project_label_name_to_label[source_label_name.casefold()].id
            else:
                raise ValueError(
                    f"Found label {source_label_name} in source labels, but this "
                    f"label is not in the project labels."
                )
            if label_id is None:
                raise ValueError(
                    f"Unable to find label id for label {source_label_name}."
                )
            source_label_name_to_project_label_id[source_label_name] = label_id
        return source_label_name_to_project_label_id

    @property
    def label_mapping(self) -> Dict[str, str]:
        """
        Return a dictionary with label names as keys and label ids as values.

        :return:
        """
        if self.annotation_reader is not None:
            if self._label_mapping is None:
                self._label_mapping = self.__get_label_mapping(self._project)
            return self._label_mapping
        else:
            raise ValueError(
                "Unable to get label mapping for this annotation client, no "
                "annotation reader has been defined."
            )

    def _upload_annotation_for_2d_media_item(
        self,
        media_item: Union[Image, VideoFrame],
        annotation_scene: Optional[AnnotationScene] = None,
    ) -> AnnotationScene:
        """
        Upload a new annotation for an image or video frame to the cluster. This will
        overwrite any current annotations for the media item.

        If an `annotation_scene` is passed, this annotation will be applied to the
        media_item.

        If `annotation_scene` is left as None and an AnnotationReader is defined for
        the AnnotationClient, this method will read the annotation for
        the media item from the AnnotationReader.

        If `annotation_scene` is left as None and no AnnotationReader is defined for
        the AnnotationClient, this method will raise a ValueError

        :param media_item: Image or VideoFrame to upload annotation for
        :param annotation_scene: Optional AnnotationScene to apply to the media_item.
            If left as None, this method will read the annotation data using the
            AnnotationReader
        :return: AnnotationScene that was uploaded
        """
        if annotation_scene is not None:
            scene_to_upload = annotation_scene.apply_identifier(
                media_identifier=media_item.identifier
            )
        else:
            if self.annotation_reader is not None:
                scene_to_upload = self._read_2d_media_annotation_from_source(
                    media_item=media_item
                )
            else:
                raise ValueError(
                    "You attempted to upload an annotation for a media item, but no "
                    "annotation data was passed directly and no annotation reader was "
                    "defined for the AnnotationClient. Therefore, the "
                    "AnnotationClient is unable to upload any annotation data."
                )
        if scene_to_upload.has_data:
            scene_to_upload.prepare_for_post()
            rest_data = AnnotationRESTConverter.to_dict(
                scene_to_upload, deidentify=False
            )
            rest_data.pop("kind")
            self.session.get_rest_response(
                url=f"{media_item.base_url}/annotations",
                method="POST",
                data=rest_data,
                include_organization_id=False,
            )
        return scene_to_upload

    def _append_annotation_for_2d_media_item(
        self, media_item: Union[Image, VideoFrame]
    ) -> AnnotationScene:
        """
        Add an annotation to the existing annotations for the `media_item`.

        :param media_item: Image or VideoFrame to append the annotation for
        :return: Returns the response of the REST endpoint to post the updated
            annotation
        """
        new_annotation_scene = self._read_2d_media_annotation_from_source(
            media_item=media_item, preserve_shape_for_global_labels=True
        )
        annotation_scene = self._get_latest_annotation_for_2d_media_item(media_item)
        if annotation_scene is None:
            logging.info(
                f"No existing annotation found for {str(media_item.type)} named "
                f"{media_item.name}"
            )
            annotation_scene = AnnotationScene(
                media_identifier=media_item.identifier,
                annotations=[],
                kind=AnnotationKind.ANNOTATION.value,
            )
        annotation_scene.extend(new_annotation_scene.annotations)

        if annotation_scene.has_data:
            rest_data = AnnotationRESTConverter.to_dict(
                annotation_scene, deidentify=False
            )
            rest_data.pop("kind", None)
            rest_data.pop("annotation_state_per_task", None)
            rest_data.pop("id", None)
            response = self.session.get_rest_response(
                url=f"{media_item.base_url}/annotations",
                method="POST",
                data=rest_data,
                include_organization_id=False,
            )
            return AnnotationRESTConverter.from_dict(response)
        else:
            return annotation_scene

    def _upload_annotations_for_2d_media_list(
        self,
        media_list: Sequence[MediaItem],
        append_annotations: bool,
        max_threads: int = 5,
    ) -> int:
        """
        Upload annotations to the server.

        :param media_list: List of images or video frames to upload the annotations
            for
        :param append_annotations: True to append annotations from the local disk to
            the existing annotations on the server, False to overwrite the server
            annotations by those on the local disk.
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: Returns the number of uploaded annotations.
        """
        if max_threads <= 0:
            # ThreadPoolExecutor will use minimum 5 threads for 1 core cpu
            # and maximum 32 threads for multi-core cpu.
            max_threads = None
        upload_count = 0
        skip_count = 0
        tqdm_prefix = "Uploading media annotations"

        def upload_annotation(media_item: MediaItem) -> None:
            nonlocal upload_count, skip_count
            try:
                if not append_annotations:
                    response = self._upload_annotation_for_2d_media_item(
                        media_item=media_item
                    )
                else:
                    response = self._append_annotation_for_2d_media_item(
                        media_item=media_item
                    )
            except GetiRequestException as error:
                skip_count += 1
                if error.status_code == 500:
                    logging.error(
                        f"Failed to upload annotation for {media_item.name}. "
                    )
                    return
                else:
                    raise error
            if response is not None:
                upload_count += 1

        t_start = time.time()
        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            with logging_redirect_tqdm(tqdm_class=tqdm):
                list(
                    tqdm(
                        executor.map(upload_annotation, media_list),
                        total=len(media_list),
                        desc=tqdm_prefix,
                    )
                )

        t_elapsed = time.time() - t_start
        if upload_count > 0:
            logging.info(
                f"Uploaded {upload_count} annotations in {t_elapsed:.1f} seconds."
            )
        if skip_count > 0:
            logging.info(
                f"Skipped {skip_count} media items, unable to upload annotations."
            )
        return upload_count

    def annotation_scene_from_rest_response(
        self, response_dict: Dict[str, Any], media_information: MediaInformation
    ) -> AnnotationScene:
        """
        Convert a dictionary with annotation data obtained from the Intel® Geti™
        /annotations rest endpoint into an annotation scene.

        :param response_dict: Dictionary containing the annotation data
        :param media_information: MediaInformation about the media item to which the
            annotation applies
        :return: AnnotationScene object corresponding to the data in the response_dict
        """
        annotation_scene = AnnotationRESTConverter.from_dict(response_dict)
        return annotation_scene

    def _get_latest_annotation_for_2d_media_item(
        self, media_item: Union[Image, VideoFrame]
    ) -> Optional[AnnotationScene]:
        """
        Retrieve the latest annotation for an image or video frame from the cluster.
        If no annotation is available, this method returns None

        :param media_item: Image or VideoFrame to retrieve the annotations for
        :return: Dictionary containing the annotations data
        """
        response = self.session.get_rest_response(
            url=f"{media_item.base_url}/annotations/latest",
            method="GET",
            include_organization_id=False,
        )
        if type(response) is Response and response.status_code == 204:
            return None
        annotation_scene = self.annotation_scene_from_rest_response(
            response, media_item.media_information
        )
        annotation_scene.resolve_label_names_and_colors(
            labels=self._project.get_all_labels()
        )
        return annotation_scene

    def _read_2d_media_annotation_from_source(
        self,
        media_item: Union[Image, VideoFrame],
        preserve_shape_for_global_labels: bool = False,
    ) -> AnnotationScene:
        """
        Retrieve the annotation for the media_item, and return it in the
        proper format to be sent to the GETi /annotations endpoint. This method uses the
        `self.annotation_reader` to get the annotation data.

        :param media_item: MediaItem to read the annotation for
        :return: Dictionary containing the annotation, in GETi format
        """
        annotation_list = self.annotation_reader.get_data(
            filename=media_item.name,
            label_name_to_id_mapping=self.label_mapping,
            media_information=media_item.media_information,
            preserve_shape_for_global_labels=preserve_shape_for_global_labels,
        )
        return AnnotationRESTConverter.from_dict(
            {
                "media_identifier": media_item.identifier,
                "annotations": annotation_list,
                "kind": AnnotationKind.ANNOTATION,
            }
        )

    def _download_annotations_for_2d_media_list(
        self,
        media_list: Union[MediaList[Image], MediaList[VideoFrame]],
        path_to_folder: str,
        append_media_uid: bool = False,
        verbose: bool = True,
        max_threads: int = 10,
    ) -> float:
        """
        Download annotations from the server to a target folder on disk.

        :param media_list: List of images or video frames to download the annotations
            for
        :param path_to_folder: Folder to save the annotations to
        :param append_media_uid: True to append the UID of a media item to the
            annotation filename (separated from the original filename by an underscore,
             i.e. '{filename}_{media_id}').
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :return: Returns the time elapsed to download the annotations, in seconds
        """
        if max_threads <= 0:
            # ThreadPoolExecutor will use minimum 5 threads for 1 core cpu
            # and maximum 32 threads for multi-core cpu.
            max_threads = None
        path_to_annotations_folder = os.path.join(path_to_folder, "annotations")
        os.makedirs(path_to_annotations_folder, exist_ok=True, mode=0o770)
        if media_list.media_type == Image:
            media_name = "image"
            media_name_plural = "images"
        elif media_list.media_type == VideoFrame:
            media_name = "video frame"
            media_name_plural = "video frames"
        else:
            raise ValueError(
                "Invalid media type found in media_list, unable to download "
                "annotations."
            )
        if verbose:
            logging.info(
                f"Starting annotation download... saving annotations for "
                f"{len(media_list)} {media_name_plural} to folder "
                f"{path_to_annotations_folder}"
            )
        t_start = time.time()
        download_count = 0
        skip_count = 0
        tqdm_prefix = f"Downloading {media_name} annotations"

        def download_annotation(media_item: Union[Image, VideoFrame]) -> None:
            nonlocal skip_count, download_count
            annotation_scene = self._get_latest_annotation_for_2d_media_item(media_item)
            if annotation_scene is None:
                log_msg = (
                    f"Unable to retrieve latest annotation for {media_name} "
                    f"{media_item.name}. Skipping this {media_name}"
                )
                if verbose:
                    logging.info(log_msg)
                else:
                    logging.debug(log_msg)
                skip_count += 1
                return
            kind = annotation_scene.kind
            if kind != AnnotationKind.ANNOTATION:
                log_msg = (
                    f"Received invalid annotation of kind {kind} for "
                    f"{media_name} with name{media_item.name}"
                )
                if verbose:
                    logging.info(log_msg)
                else:
                    logging.debug(log_msg)
                skip_count += 1
                return
            export_data = AnnotationRESTConverter.to_dict(annotation_scene)

            base_media_item_name = os.path.basename(media_item.name)
            filename = base_media_item_name + ".json"
            if append_media_uid:
                if isinstance(media_item, Image):
                    filename = f"{base_media_item_name}_{media_item.id}.json"
                elif isinstance(media_item, VideoFrame):
                    if media_item.video_name is not None:
                        base_video_name = os.path.basename(media_item.video_name)
                        filename = (
                            f"{base_video_name}_"
                            f"{media_item.media_information.video_id}_frame_"
                            f"{media_item.media_information.frame_index}.json"
                        )
                    else:
                        video_name = base_media_item_name.split("_frame_")[0]
                        filename = (
                            f"{video_name}_"
                            f"{media_item.media_information.video_id}_frame_"
                            f"{media_item.media_information.frame_index}.json"
                        )
                else:
                    raise TypeError(
                        f"Received invalid media item of type {type(media_item)}."
                    )

            annotation_path = os.path.join(path_to_annotations_folder, filename)

            with open(annotation_path, "w") as f:
                json.dump(export_data, f, indent=4)
            download_count += 1

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            with logging_redirect_tqdm(tqdm_class=tqdm):
                list(
                    tqdm(  # List unwraps the generator
                        executor.map(download_annotation, media_list),
                        total=len(media_list),
                        desc=tqdm_prefix,
                    )
                )

        t_elapsed = time.time() - t_start
        if download_count > 0:
            msg = (
                f"Downloaded {download_count} annotations to folder "
                f"{path_to_annotations_folder} in {t_elapsed:.1f} seconds."
            )
        else:
            msg = "No annotations were downloaded."
        if skip_count > 0:
            msg = (
                msg + f" Was unable to retrieve annotations for {skip_count} "
                f"{media_name_plural}, these {media_name_plural} were skipped."
            )
        if verbose:
            logging.info(msg)
        return t_elapsed



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import ClassVar, Dict, Optional

import attr

from geti_sdk.data_models.utils import attr_value_serializer, str_to_media_type


@attr.define
class MediaIdentifier:
    """
    Representation of media identification data as output by the Intel® Geti™
    /annotations REST endpoints.

    :var type: Type of the media to which the annotation belongs
    """

    type: str = attr.field(converter=str_to_media_type)

    def to_dict(self) -> Dict[str, str]:
        """
        Return a dictionary form of the MediaIdentifier instance.

        :return: Dictionary containing the media identifier data
        """
        return attr.asdict(self, value_serializer=attr_value_serializer)


@attr.define
class ImageIdentifier(MediaIdentifier):
    """
    Representation of image identification data used by the Intel® Geti™ /annotations
    endpoints. This object uniquely identifies an Image on the Intel® Geti™ server.

    :var image_id: unique database ID of the image
    """

    _identifier_fields: ClassVar[str] = ["image_id"]

    image_id: str


@attr.define
class VideoFrameIdentifier(MediaIdentifier):
    """
    Representation of video frame identification data used by the Intel® Geti™
    /annotations endpoints. This object uniquely identifies a VideoFrame on the
    Intel® Geti™ server.

    :var frame_index: Index of the video frame in the full video
    :var video_id: unique database ID of the video to which the frame belongs
    :var key_index: Index of the key frame in the video
    """

    _identifier_fields: ClassVar[str] = ["video_id"]

    frame_index: int
    video_id: str
    key_index: Optional[int] = None  # Added in Geti 2.5


@attr.define
class VideoIdentifier(MediaIdentifier):
    """
    Representation of video identification data used by the Intel® Geti™ /annotations
    endpoints. This object uniquely identifiers a Video on the Intel® Geti™ server.

    :var video_id: unique database ID of the video
    """

    _identifier_fields: ClassVar[str] = ["video_id"]

    video_id: str



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import os
import time
from typing import Dict, List, Optional, Union

from geti_sdk.data_models import (
    Algorithm,
    FullConfiguration,
    GlobalConfiguration,
    Project,
    Task,
    TaskConfiguration,
)
from geti_sdk.data_models.configurable_parameter_group import PARAMETER_TYPES
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_converters import ConfigurationRESTConverter
from geti_sdk.utils import get_supported_algorithms


class ConfigurationClient:
    """
    Class to manage configuration for a certain project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        project_id = project.id
        self.project = project
        self.workspace_id = workspace_id
        self.task_ids = [task.id for task in project.get_trainable_tasks()]
        self.base_url = (
            f"workspaces/{workspace_id}/projects/{project_id}/" f"configuration"
        )
        self.supported_algos = get_supported_algorithms(
            rest_session=session, project=project, workspace_id=workspace_id
        )

        # Query the project status to make sure that the project is loaded. Then we
        # can safely fetch the configuration later on, even for newly created projects
        self.session.get_rest_response(
            url=f"workspaces/{workspace_id}/projects/{project_id}/status", method="GET"
        )
        # Hack: Wait for some time to make sure that the configurations are
        # initialized properly
        time.sleep(1)

    def get_task_configuration(
        self, task_id: str, algorithm_name: Optional[str] = None
    ) -> TaskConfiguration:
        """
        Get the configuration for the task with id `task_id`.

        :param task_id: ID of the task to get configurations for
        :param algorithm_name: Optional name of the algorithm to get configuration for.
            If an algorithm name is passed, the returned TaskConfiguration will contain
            only the hyper parameters for that algorithm, and won't hold any
            component parameters
        :return: TaskConfiguration holding all component parameters and hyper parameters
            for the task
        """
        url = f"{self.base_url}/task_chain/{task_id}"
        if algorithm_name is not None:
            url += f"?algorithm_name={algorithm_name}"
        config_data = self.session.get_rest_response(url=url, method="GET")
        return ConfigurationRESTConverter.task_configuration_from_dict(config_data)

    def get_global_configuration(self) -> GlobalConfiguration:
        """
        Get the project-wide configurable parameters.

        :return: GlobalConfiguration instance holding the configurable parameters for
            all project-wide components
        """
        config_data = self.session.get_rest_response(
            url=f"{self.base_url}/global", method="GET"
        )
        return ConfigurationRESTConverter.global_configuration_from_rest(config_data)

    def _set_task_configuration(self, task_id: str, config: dict):
        """
        Update the configuration for a task.

        :param task_id: ID of the task to set the configuration for
        :param config: Dictionary containing the updated configuration values
        :return: Response of the configuration POST endpoint.
        """
        response = self.session.get_rest_response(
            url=f"{self.base_url}/task_chain/{task_id}", method="POST", data=config
        )
        return response

    def set_project_auto_train(self, auto_train: bool = False) -> None:
        """
        Set the `auto_train` parameter for all tasks in the project.

        :param auto_train: True to enable auto_training, False to disable
        """
        for task_id in self.task_ids:
            config = self.get_task_configuration(task_id=task_id)
            config_data = config.set_parameter_value("auto_training", value=auto_train)
            self._set_task_configuration(task_id=task_id, config=config_data)

    def set_project_num_iterations(self, value: int = 50):
        """
        Set the number of iterations to train for each task in the project.

        :param value: Number of iterations to set
        """
        iteration_names = ["num_iters", "max_num_epochs"]
        for task_id in self.task_ids:
            config = self.get_task_configuration(task_id=task_id)
            parameter: Optional[PARAMETER_TYPES] = None
            for parameter_name in iteration_names:
                parameter = config.get_parameter_by_name(parameter_name)
                if parameter is not None:
                    self._set_task_configuration(
                        task_id=task_id,
                        config=config.set_parameter_value(parameter_name, value=value),
                    )
                    break
            if parameter is None:
                raise ValueError(
                    f"No iteration parameters were found for task {config.task_title}. "
                    f"Unable to set number of iterations"
                )

    def set_project_parameter(
        self,
        parameter_name: str,
        value: Union[bool, str, float, int],
        parameter_group_name: Optional[str] = None,
    ):
        """
        Set the value for a parameter with `parameter_name` that lives in the
        group `parameter_group_name`. The parameter is set for all tasks in the project

        The `parameter_group_name` can be left as None, in that case this method will
        attempt to determine the appropriate parameter group automatically.

        :param parameter_name: Name of the parameter
        :param parameter_group_name: Optional name of the parameter group name to
            which the parameter belongs. If left as None (the default), this method will
            attempt to determine the correct parameter group automatically, if needed.
        :param value: Value to set for the parameter
        """
        for index, task_id in enumerate(self.task_ids):
            config = self.get_task_configuration(task_id)
            config_data = config.set_parameter_value(
                parameter_name=parameter_name,
                value=value,
                group_name=parameter_group_name,
            )
            self._set_task_configuration(task_id=task_id, config=config_data)

    def get_full_configuration(self) -> FullConfiguration:
        """
        Return the full configuration for a project (for both global and task_chain).

        :return: FullConfiguration object holding the global and task chain
            configuration
        """
        data = self.session.get_rest_response(url=self.base_url, method="GET")
        return ConfigurationRESTConverter.full_configuration_from_rest(data)

    def get_for_task_and_algorithm(self, task: Task, algorithm: Algorithm):
        """
        Get the hyper parameters for a specific task and algorithm.

        :param task: Task to get hyper parameters for
        :param algorithm: Algorithm to get hyper parameters for
        :return: TaskConfiguration holding only the model hyper parameters for the
            specified algorithm
        """
        if algorithm not in self.supported_algos.get_by_task_type(task.type):
            raise ValueError(
                f"The requested algorithm '{algorithm.name}' is not "
                f"supported for a task of type '{task.type}'. Unable to retrieve "
                f"configuration."
            )
        return self.get_task_configuration(
            task_id=task.id, algorithm_name=algorithm.model_template_id
        )

    def download_configuration(self, path_to_folder: str) -> FullConfiguration:
        """
        Retrieve the full configuration for a project from the cluster and save it to
        a file `configuration.json` in the folder specified at `path_to_folder`.

        :param path_to_folder: Folder to save the configuration to
        :return:
        """
        config = self.get_full_configuration()
        config_data = ConfigurationRESTConverter.configuration_to_minimal_dict(config)
        os.makedirs(path_to_folder, exist_ok=True, mode=0o770)
        configuration_path = os.path.join(path_to_folder, "configuration.json")
        with open(configuration_path, "w") as file:
            json.dump(config_data, file, indent=4)
        logging.info(
            f"Project parameters for project '{self.project.name}' were saved to file "
            f"{configuration_path}."
        )
        return config

    def apply_from_object(
        self, configuration: FullConfiguration
    ) -> Optional[FullConfiguration]:
        """
        Attempt to apply the configuration values passed in as `configuration` to
        the project managed by this instance of the ConfigurationClient.

        :param configuration: FullConfiguration to be applied
        :return:
        """
        global_config = configuration.global_
        global_config.apply_identifiers(self.workspace_id, self.project.id)

        project_tasks = self.project.get_trainable_tasks()
        if len(project_tasks) != len(configuration.task_chain):
            raise ValueError(
                f"Structure of the configuration in: '{configuration}' does not match "
                f"that of the project. Unable to set configuration"
            )
        for task, task_config in zip(project_tasks, configuration.task_chain):
            current_task_config = self.get_task_configuration(task_id=task.id)
            model_storage_ids = [
                config.entity_identifier.model_storage_id
                for config in current_task_config.model_configurations
            ]

            task_config.apply_identifiers(
                workspace_id=self.workspace_id,
                project_id=self.project.id,
                task_id=task.id,
                model_storage_id=model_storage_ids[0],
            )
        data = ConfigurationRESTConverter.configuration_to_minimal_dict(
            configuration=configuration, deidentify=False
        )
        try:
            result = self.session.get_rest_response(
                url=self.base_url, method="POST", data=data
            )
        except GetiRequestException:
            failed_parameters: List[Dict[str, str]] = []
            global_config = configuration.global_
            task_chain_config = configuration.task_chain
            for parameter in global_config:
                config_data = global_config.set_parameter_value(
                    parameter.name, parameter.value
                )
                try:
                    self.session.get_rest_response(
                        url=f"{self.base_url}/global", method="POST", data=config_data
                    )
                except GetiRequestException:
                    failed_parameters.append({"global": parameter.name})
            for task_config in task_chain_config:
                for parameter in task_config:
                    config_data = task_config.set_parameter_value(
                        parameter.name, parameter.value
                    )
                    try:
                        self._set_task_configuration(
                            task_id=task_config.task_id, config=config_data
                        )
                    except GetiRequestException:
                        failed_parameters.append(
                            {task_config.task_title: parameter.name}
                        )
            logging.warning(
                f"Setting configuration failed for the following parameters: "
                f"{failed_parameters}. All other parameters were set successfully."
            )
            result = None
        if result:
            return configuration
        else:
            return None

    def apply_from_file(
        self, path_to_folder: str, filename: Optional[str] = None
    ) -> Optional[FullConfiguration]:
        """
        Attempt to apply a configuration from a file on disk. The
        parameter `path_to_folder` is mandatory and should point to the folder in which
        the configuration file to upload lives. The parameter `filename` is optional,
        when left as `None` this method will look for a file `configuration.json` in
        the specified folder.

        :param path_to_folder: Path to the folder in which the configuration file to
            apply lives
        :param filename: Optional filename for the configuration file to apply
        :return:
        """
        if filename is None:
            filename = "configuration.json"
        path_to_config = os.path.join(path_to_folder, filename)
        if not os.path.isfile(path_to_config):
            raise ValueError(
                f"Unable to find configuration file at {path_to_config}. Please "
                f"provide a valid path to the folder holding the configuration data."
            )
        with open(path_to_config, "r") as file:
            data = json.load(file)
        config = ConfigurationRESTConverter.full_configuration_from_rest(data)
        return self.apply_from_object(config)

    def set_configuration(
        self,
        configuration: Union[FullConfiguration, GlobalConfiguration, TaskConfiguration],
    ):
        """
        Set the configuration for the project. This method accepts either a
        FullConfiguration, TaskConfiguration or GlobalConfiguration object

        :param configuration: Configuration to set
        :return:
        """
        if isinstance(configuration, FullConfiguration):
            full_configuration = configuration
        elif isinstance(configuration, TaskConfiguration):
            full_configuration = self.get_full_configuration()
            if configuration.task_id is None:
                raise ValueError(
                    "Cannot set a TaskConfiguration without a task_id. Please make "
                    "sure to set a valid task_id."
                )
            # Find index of configuration in the task chain
            task_index: Optional[int] = None
            for ti, task_configuration in enumerate(full_configuration.task_chain):
                if task_configuration.task_id == configuration.task_id:
                    task_index = ti
                    break
            if task_index is None:
                raise ValueError(
                    f"Unable to find task with id {configuration.task_id} in the "
                    f"project. Please make sure that the task_id for the configuration "
                    f"you have provided is valid"
                )
            full_configuration.task_chain[task_index] = configuration
        elif isinstance(configuration, GlobalConfiguration):
            full_configuration = self.get_full_configuration()
            full_configuration.global_ = configuration
        else:
            raise TypeError(
                f"Invalid configuration of type '{type(configuration)}' received. "
                f"Unable to set configuration."
            )
        self.apply_from_object(full_configuration)

    def get_for_model(self, task_id: str, model_id: str) -> TaskConfiguration:
        """
        Get the hyper parameters for the model with id `model_id`. Note that the model
        has to be trained within the task with id `task_id` in order for the
        parameters to be retrieved successfully.

        :param task_id: ID of the task to get configurations for
        :param model_id: ID of the model to get the hyper parameters for
        :return: TaskConfiguration holding all hyper parameters for the model
        """
        url = f"{self.base_url}/task_chain/{task_id}?model_id={model_id}"
        config_data = self.session.get_rest_response(url=url, method="GET")
        return ConfigurationRESTConverter.task_configuration_from_dict(config_data)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
import tempfile
from typing import Union

import cv2
import numpy as np

from geti_sdk.data_models import Image, Prediction
from geti_sdk.utils import show_image_with_annotation_scene


def add_text_to_top_of_image(image: np.ndarray, text: str) -> np.ndarray:
    """
    Adds the specified text to the center top of the image.

    :param image: Image to add text to
    :param text: String to display
    :return: image with text added to the top center
    """
    font = cv2.FONT_HERSHEY_SIMPLEX
    font_scale = 1
    thickness = 1
    text_size, baseline = cv2.getTextSize(text, font, font_scale, thickness)
    origin = [int(0.5 * (image.shape[1] - text_size[0])), 2 + text_size[1]]
    cv2.putText(
        image,
        text,
        org=origin,
        fontFace=font,
        fontScale=font_scale,
        color=(0, 0, 0),
        thickness=thickness,
    )
    return image


def plot_predictions_side_by_side(
    image: Union[Image, np.ndarray],
    prediction_1: Prediction,
    prediction_2: Prediction,
    filepath: str,
):
    """
    Plots two predictions next to each other

    :param image: Image on which to plot the predictions
    :param prediction_1: First Prediction for image
    :param prediction_2: Second Prediction for image
    :param filepath: Path to file in which to save the resulting image
    :return: resulting image
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        im_1_pred_path = os.path.join(temp_dir, "im_1.jpg")
        im_2_pred_path = os.path.join(temp_dir, "im_2.jpg")
        show_image_with_annotation_scene(
            image=image, annotation_scene=prediction_1, filepath=im_1_pred_path
        )
        show_image_with_annotation_scene(
            image=image, annotation_scene=prediction_2, filepath=im_2_pred_path
        )
        im1 = cv2.imread(im_1_pred_path)
        im2 = cv2.imread(im_2_pred_path)

    add_text_to_top_of_image(im1, "local prediction")
    add_text_to_top_of_image(im2, "online prediction")
    separator = np.zeros((im1.shape[0], 1, im1.shape[2]))
    final_im = np.concatenate((im1, separator, im2), axis=1)
    cv2.imwrite(filepath, final_im)
    return final_im



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""The package provides the Visualizer class for models predictions visualization."""


from os import PathLike
from typing import List, Optional, Union

import cv2
import numpy as np
from IPython.display import display
from PIL import Image

from geti_sdk.data_models.annotation_scene import AnnotationScene
from geti_sdk.data_models.containers.media_list import MediaList
from geti_sdk.data_models.media import VideoFrame
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.prediction_visualization.shape_drawer import ShapeDrawer


class Visualizer:
    """
    Visualize the predicted output by drawing the annotations on the input image.

    :example:
        >>> prediction = deployment.infer(image)
        >>> visualizer = Visualizer()
        >>> output_image = visualizer.draw(image, prediction )
        >>> Display(output_image)

    """

    def __init__(
        self,
        window_name: Optional[str] = None,
        show_labels: bool = True,
        show_confidence: bool = True,
        show_count: bool = False,
        is_one_label: bool = False,
    ) -> None:
        """
        Initialize the Visualizer.

        :param window_name: Name of the window to be shown
        :param show_labels: Show labels on the output image
        :param show_confidence: Show confidence on the output image
        :param show_count: Show count of the shapes on the output image
        :param is_one_label: Show only one label on the output image
        """
        self.window_name = "Window" if window_name is None else window_name
        self.shape_drawer = ShapeDrawer(
            show_count, is_one_label, show_labels, show_confidence
        )

    def draw(
        self,
        image: np.ndarray,
        annotation: AnnotationScene,
        fill_shapes: bool = True,
        confidence_threshold: Optional[float] = None,
        meta: Optional[dict] = None,
    ) -> np.ndarray:
        """
        Draw annotations on the image.

        :param image: Input image in RGB format
        :param annotation: Annotations to be drawn on the input image
        :param meta: Optional meta information
        :param fill_shapes: Fill shapes with color
        :param confidence_threshold: Confidence threshold to filter annotations.
            Must be in range [0, 1].
        :return: Output image with annotations in RGB format
        """
        if confidence_threshold is not None:
            annotation = annotation.filter_by_confidence(confidence_threshold)
        result = self.shape_drawer.draw(
            image.copy(), annotation, labels=[], fill_shapes=fill_shapes
        )
        return result

    def explain_label(
        self,
        image: np.ndarray,
        prediction: Prediction,
        label_name: str,
        opacity: float = 0.5,
        show_predictions: bool = True,
    ) -> np.ndarray:
        """
        Draw saliency map overlay on the image.

        :param image: Input image in RGB format
        :param prediction: Prediction object containing saliency maps
        :param label_name: Label name to be explained
        :param opacity: Opacity of the saliency map overlay
        :param show_predictions: Show predictions for the label on the output image
        :return: Output image with saliency map overlay in RGB format
        """
        saliency_map = None
        for pred_map in prediction.maps:
            if pred_map.type == "saliency map":
                saliency_map = pred_map.data
                break
        if saliency_map is None:
            raise ValueError("Prediction does not contain saliency maps")
        if label_name not in saliency_map:
            raise ValueError(
                f"Saliency map for label {label_name} is not found in the prediction."
            )
        # Accessing the saliency map for the label
        saliency_map = saliency_map[label_name]
        if saliency_map.shape[:2] != image.shape[:2]:
            saliency_map = cv2.resize(
                saliency_map,
                (image.shape[1], image.shape[0]),
                interpolation=cv2.INTER_CUBIC,
            )
        # Visualizing the saliency map
        overlay = cv2.applyColorMap(saliency_map, cv2.COLORMAP_JET)
        overlay = cv2.cvtColor(overlay, cv2.COLOR_BGR2RGB)
        result = cv2.addWeighted(image, 1 - opacity, overlay, opacity, 0)
        if show_predictions:
            filtered_prediction = prediction.filter_annotations([label_name])
            result = self.draw(result, filtered_prediction, fill_shapes=False)
        return result

    @staticmethod
    def save_image(image: np.ndarray, output_path: PathLike) -> None:
        """
        Save the image to the output path.

        :param image: Image in RGB format to be saved
        :param output_path: Path to save the image
        """
        bgr_image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        cv2.imwrite(output_path, bgr_image)

    @staticmethod
    def save_video(
        video_frames: MediaList[VideoFrame],
        annotation_scenes: List[Union[AnnotationScene, Prediction]],
        output_path: PathLike,
        fps: float = 1,
    ) -> None:
        """
        Save the video to the output path.

        :param video_frames: List of video frames
        :param annotation_scenes: List of annotation scenes to be drawn on the video frames
        :param output_path: Path to save the image
        """
        out_writer = cv2.VideoWriter(
            filename=f"{output_path}",
            fourcc=cv2.VideoWriter_fourcc("M", "J", "P", "G"),
            fps=fps,
            frameSize=(
                video_frames[0].media_information.width,
                video_frames[0].media_information.height,
            ),
        )
        for frame, annotation in zip(video_frames, annotation_scenes):
            out_writer.write(frame)

    @staticmethod
    def show_in_notebook(image: np.ndarray) -> None:
        """
        Show the image in the Jupyter notebook.

        :param image: Image to be shown in RGB format
        """
        display(Image.fromarray(image))

    def show_window(self, image: np.ndarray) -> None:
        """
        Show result image.

        :param image: Image to be shown (in RGB order).
        """
        pil_image = Image.fromarray(image)
        pil_image.show(title=self.window_name)

    def is_quit(self, delay: int = 1) -> bool:
        """Check user wish to quit."""
        return ord("q") == cv2.waitKey(delay)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import glob
import json
import logging
import os
import warnings
from random import sample
from typing import Any, Dict, List, Optional, Union

from geti_sdk.data_models import Annotation, TaskType
from geti_sdk.data_models.media import MediaInformation
from geti_sdk.data_models.shapes import Rectangle
from geti_sdk.rest_converters import AnnotationRESTConverter
from geti_sdk.rest_converters.annotation_rest_converter import (
    NormalizedAnnotationRESTConverter,
)

from .base_annotation_reader import AnnotationReader


class GetiAnnotationReader(AnnotationReader):
    """
    AnnotationReader for loading annotation files in Intel® Geti™ format.
    """

    def __init__(
        self,
        base_data_folder: str,
        annotation_format: str = ".json",
        task_type: Optional[Union[TaskType, str]] = None,
        label_names_to_include: Optional[List[str]] = None,
        anomaly_reduction: bool = False,
    ):
        """
        :param base_data_folder: Path to the folder containing the annotations
        :param annotation_format: Extension of the annotation files. Defaults to ".json"
        :param task_type: Optional task type to prepare the annotations for. Can also
            be specified later.
        :param label_names_to_include: Names of the labels that should be included
            when reading annotation data. This can be used to filter the annotations
            for certain labels.
        :param anomaly_reduction: True to reduce all anomaly tasks to the single anomaly task.
            This is done in accordance with the Intel Geti 2.5 Anomaly Reduction effort.
            All pixel level annotations are converted to full rectangles. All anomaly tasks
            are mapped to th new "Anomaly Detection" task wich corresponds to the old "Anomaly Classification".
        """
        if annotation_format != ".json":
            raise ValueError(
                f"Annotation format {annotation_format} is currently not"
                f" supported by the GetiAnnotationReader"
            )
        super().__init__(
            base_data_folder=base_data_folder,
            annotation_format=annotation_format,
            task_type=task_type,
            anomaly_reduction=anomaly_reduction,
        )
        self._label_names_to_include = label_names_to_include
        self._normalized_annotations = self._has_normalized_annotations()

    def _get_label_names(self, all_labels: List[str]) -> List[str]:
        """
        Return the labels for the task type the annotation reader is currently set to.

        :param all_labels: List of all label names in the project
        """
        if self._label_names_to_include is None:
            logging.info("No label mapping defined, including all labels")
            labels = all_labels
        else:
            labels = [
                label for label in all_labels if label in self._label_names_to_include
            ]
        return labels

    def _get_raw_annotation_data(self, filename: str) -> Dict[str, Any]:
        """
        Read the annotation data from the file at `filename`

        :param filename: Name of the annotation file to read
        :return: Dictionary holding the annotation data
        """
        filepath = glob.glob(
            os.path.join(self.base_folder, f"{filename}{self.annotation_format}"),
        )
        if len(filepath) > 1:
            warnings.warn(
                f"Multiple matching annotation files found for image with "
                f"name {filename}. Skipping this image..."
            )
            data = {"annotations": []}
        elif len(filepath) == 0:
            logging.info(
                f"No matching annotation file found for image with name {filename}."
                f" Skipping this image..."
            )
            data = {"annotations": []}
        else:
            filepath = filepath[0]
            with open(filepath, "r") as f:
                data = json.load(f)
        return data

    def get_data(
        self,
        filename: str,
        label_name_to_id_mapping: dict,
        media_information: MediaInformation,
        preserve_shape_for_global_labels: bool = False,
    ) -> List[Annotation]:
        """
        Return the annotation data for the dataset item corresponding to `filename`.

        :param filename: name of the item to get the annotation data for.
        :param label_name_to_id_mapping: mapping of label name to label id.
        :param media_information: MediaInformation object containing information
            (e.g. width, height) about the media item to upload the annotation for
        :param preserve_shape_for_global_labels: False to convert shapes for global
            tasks to full rectangles (required for classification like tasks in
            Intel® Geti™ projects), True to preserve such shapes. This parameter
            should be:

             - False when uploading annotations to a single task project
             - True when uploading annotations for a classification like task,
                following a local task in a task chain project.

        :return: List of Annotation objects containing all annotations for the given
            dataset item.
        """
        data = self._get_raw_annotation_data(filename=filename)

        new_annotations = []
        for annotation in data["annotations"]:
            if self._normalized_annotations:
                annotation_object = (
                    NormalizedAnnotationRESTConverter.normalized_annotation_from_dict(
                        annotation,
                        image_width=media_information.width,
                        image_height=media_information.height,
                    )
                )
            else:
                annotation_object = AnnotationRESTConverter.annotation_from_dict(
                    annotation
                )
            for label in annotation_object.labels:
                label.id = label_name_to_id_mapping[label.name]
            for label_dict in annotation["labels"]:
                if self.task_type is not None:
                    if label_dict["name"] not in self._get_label_names(
                        list(label_name_to_id_mapping.keys())
                    ):
                        annotation_object.pop_label_by_name(
                            label_name=label_dict["name"]
                        )
            new_annotations.append(annotation_object)
            if (
                self.anomaly_reduction
                and annotation_object.labels[0].name.lower() == "anomalous"
            ):
                # Part of anomaly task reduction in Intel Geti 2.5 -> all anomaly tasks combined into one.
                # Intel Geti now only accepts full rectangles for anomaly tasks.
                new_annotations = [
                    Annotation(
                        labels=[annotation_object.labels[0]],
                        shape=Rectangle.generate_full_box(
                            image_width=media_information.width,
                            image_height=media_information.height,
                        ),
                    )
                ]
                break
        return new_annotations

    def get_all_label_names(self) -> List[str]:
        """
        Retrieve the unique label names for all annotations in the annotation folder

        :return: List of label names
        """
        logging.info(f"Reading annotation files in folder {self.base_folder}...")
        unique_label_names = []
        annotation_files = glob.glob(
            os.path.join(self.base_folder, f"*{self.annotation_format}"),
        )
        if len(annotation_files) == 0:
            raise ValueError(
                f"No valid annotation files were found in folder {self.base_folder}"
            )
        for annotation_file in annotation_files:
            with open(annotation_file, "r") as f:
                data = json.load(f)
            annotations = data.get("annotations", None)
            if annotations is None:
                raise ValueError(
                    f"Annotation file '{annotation_file}' does not contain any "
                    f"annotations. Please make sure that this is a valid "
                    f"annotation file."
                )
            for annotation in annotations:
                labels = [label["name"] for label in annotation["labels"]]
                for label in labels:
                    if label not in unique_label_names:
                        unique_label_names.append(label)
        return unique_label_names

    def _has_normalized_annotations(self) -> bool:
        """
        Check if the annotation files belonging to this annotation reader are normalized
        """
        filenames = self.get_data_filenames()
        n_sample = min(len(filenames), 50)
        if n_sample == 50:
            list_to_check = sample(filenames, n_sample)
        else:
            list_to_check = filenames

        NORMALIZED_KEY = "normalized"
        PIXEL_KEY = "pixel"

        annotation_stats = {NORMALIZED_KEY: 0, PIXEL_KEY: 0}
        for filename in list_to_check:
            data = self._get_raw_annotation_data(filename=filename)
            for annotation_dict in data["annotations"]:
                annotation_object = AnnotationRESTConverter.annotation_from_dict(
                    annotation_dict
                )
                shape = annotation_object.shape
                x_max, y_max = shape.x_max, shape.y_max
                if x_max <= 1 and y_max <= 1:
                    annotation_stats[NORMALIZED_KEY] += 1
                else:
                    annotation_stats[PIXEL_KEY] += 1

        if annotation_stats[NORMALIZED_KEY] == 0:
            return False
        elif annotation_stats[PIXEL_KEY] == 0:
            logging.info(
                "Legacy annotation format detected. The annotations you are trying to "
                "upload were most likely downloaded from a pre-production version of "
                "the Intel Geti software. They will be converted to the latest "
                "annotation format upon upload to the Intel Geti platform. "
            )
            return True
        else:
            raise ValueError(
                f"The annotation directory '{self.base_folder}' contains both "
                f"normalized ({annotation_stats[NORMALIZED_KEY]} shapes) and "
                f"non-normalized ({annotation_stats[PIXEL_KEY]} shapes) objects. "
                f"Unable to parse annotation data."
            )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from datetime import datetime
from typing import ClassVar, List, Optional, Union

import attr

from geti_sdk.data_models.algorithms import Algorithm
from geti_sdk.data_models.model import Model, ModelPurgeInfo
from geti_sdk.data_models.performance import Performance
from geti_sdk.data_models.utils import str_to_datetime


@attr.define
class ModelSummary:
    """
    Representation of a Model on the Intel® Geti™ platform, containing only the
    minimal information about the model.

    :var name: Name of the model
    :var creation_date: Creation date of the model
    :var version: Model version
    :var score: Score that was achieved upon evaluation of the model on the test set
    :var active_model: True if this model was the active model for the project it was
        created in, False if it was not the active model
    :var id: Unique database ID of the model
    :var model_storage_id: Unique database ID of the model storage (also referred to
        as model group) that this model belongs to
    :var label_schema_in_sync: Boolean indicating whether the labels of the model are
        matching with the latest project labels
    """

    _identifier_fields: ClassVar[List[str]] = ["id", "model_storage_id"]

    name: str
    creation_date: str = attr.field(converter=str_to_datetime)
    score_up_to_date: Optional[bool] = None  # Deprecated in Geti 2.6
    purge_info: Optional[ModelPurgeInfo] = None
    size: Optional[int] = None
    version: Optional[int] = None  # 'version' is removed in v1.1
    score: Optional[float] = attr.field(default=None)  # 'score' is removed in v1.1
    performance: Optional[Performance] = None
    active_model: bool = attr.field(default=False)
    id: Optional[str] = attr.field(default=None, repr=False)
    model_storage_id: Optional[str] = attr.field(default=None, repr=False)
    label_schema_in_sync: Optional[bool] = attr.field(default=None)  # Added in Geti 1.1


@attr.define(slots=False)
class ModelGroup:
    """
    Representation of a ModelGroup on the Intel® Geti™ server. A model group is a
    collection of models that all share the same neural network architecture, but may
    have been trained with different training datasets or hyper parameters.
    """

    _identifier_fields: ClassVar[List[str]] = ["id", "task_id"]

    name: str
    model_template_id: str
    models: List[ModelSummary] = attr.field(repr=False)
    task_id: Optional[str] = attr.field(default=None)
    id: Optional[str] = attr.field(default=None)
    learning_approach: Optional[str] = attr.field(default=None)  # Added in Geti v2.5
    lifecycle_stage: Optional[str] = attr.field(default=None)  # Added in Geti v2.6

    def __attrs_post_init__(self) -> None:
        """
        Initialize private attributes.
        """
        self._algorithm: Optional[Algorithm] = None

    @property
    def has_trained_models(self) -> bool:
        """
        Return True if the ModelGroup contains at least one trained model.

        :return: True if the model group holds at least one trained model, False
            otherwise
        """
        trained_models = [
            model
            for model in self.models
            if (model.performance is not None or model.score is not None)
        ]
        return len(trained_models) > 0

    def get_latest_model(self) -> Optional[ModelSummary]:
        """
        Return the latest model in the model group.

        :return: summary information of the most recently model in the model group
        """
        if not self.has_trained_models:
            return None
        creation_dates = [model.creation_date for model in self.models]
        return self.get_model_by_creation_date(max(creation_dates))

    def get_model_by_version(self, version: int) -> ModelSummary:
        """
        Return the model with version `version` in the model group. If no model with
        the version is found, this method raises a ValueError.

        :param version: Number specifying the desired model version
        :return: ModelSummary instance with the specified version, if any
        """
        if not self.has_trained_models:
            return None
        try:
            model = next((model for model in self.models if model.version == version))
        except StopIteration:
            raise ValueError(
                f"Model with version {version} does not exist in model group {self}"
            )
        return model

    def get_model_by_creation_date(self, creation_date: datetime) -> ModelSummary:
        """
        Return the model created on `creation_date` in the model group. If no model
        by that date is found, this method raises a ValueError

        :param creation_date: Datetime object representing the desired creation_date
        :return: ModelSummary instance with the specified creation_date, if any
        """
        if not self.has_trained_models:
            return None
        try:
            model = next(
                (model for model in self.models if model.creation_date == creation_date)
            )
        except StopIteration:
            raise ValueError(
                f"Model with creation date {creation_date} does not exist in model "
                f"group {self}"
            )
        return model

    @property
    def algorithm(self) -> Optional[Algorithm]:
        """
        Return the details for the algorithm corresponding to the ModelGroup
        This property will return None unless the `get_algorithm_details` method is
        called to retrieve the algorithm information from the Intel® Geti™ server

        :return: Algorithm details, if available
        """
        return self._algorithm

    @algorithm.setter
    def algorithm(self, algorithm: Algorithm) -> None:
        """
        Set the algorithm details for this ModelGroup instance.

        :param algorithm: Algorithm information to set
        """
        self._algorithm = algorithm

    def contains_model(self, model: Union[ModelSummary, Model]) -> bool:
        """
        Return True if the model group contains the `model`.

        :param model: Model or ModelSummary object
        :return: True if the group contains the model, False otherwise
        """
        return model.id in [model.id for model in self.models]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
"""Triggers that can be used in post inference hooks"""
from .always_trigger import AlwaysTrigger
from .confidence_trigger import ConfidenceTrigger
from .empty_label_trigger import EmptyLabelTrigger
from .label_trigger import LabelTrigger
from .object_count_trigger import ObjectCountTrigger

__all__ = [
    "AlwaysTrigger",
    "ConfidenceTrigger",
    "LabelTrigger",
    "EmptyLabelTrigger",
    "ObjectCountTrigger",
]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
"""
Credit system-related entities
"""

from copy import deepcopy
from pprint import pformat
from typing import Any, Dict, Optional

import attr

from geti_sdk.data_models.enums import SubscriptionStatus
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    deidentify,
    remove_null_fields,
    str_to_datetime,
    str_to_enum_converter,
)


@attr.define
class CreditBalance:
    """
    Representation of the Credit Balance in Intel Geti
    """

    incoming: int
    available: int
    blocked: Optional[int] = None


@attr.define
class CreditAccount:
    """
    Representation of the Credit Account in Intel Geti
    """

    _identifier_fields = [
        "id",
        "organization_id",
    ]
    id: str
    organization_id: str
    name: str
    balance: CreditBalance
    created: str = attr.field(converter=str_to_datetime)
    updated: str = attr.field(converter=str_to_datetime)
    expires: Optional[str] = attr.field(
        default=None, converter=str_to_datetime
    )  # renewable account doesn't have an expiration timestamp by default
    renewal_day_of_month: Optional[int] = None
    # renewable quota, for the welcoming one-off account it's None
    renewable_amount: Optional[int] = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the object.

        :return:
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    @property
    def overview(self) -> str:
        """
        Return a string that represents an overview of the model.

        :return:
        """
        deidentified = deepcopy(self)
        deidentified.deidentify()
        overview_dict = deidentified.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)

    def deidentify(self) -> None:
        """
        Remove unique database IDs from the base object.
        """
        deidentify(self)


@attr.define
class Subscription:
    """
    Representation of the Subscription in Intel Geti
    """

    _identifier_fields = [
        "id",
        "product_id",
        "organization_id",
        "workspace_id",
    ]
    id: str
    product_id: str
    organization_id: str
    workspace_id: str
    status: SubscriptionStatus = attr.field(
        converter=str_to_enum_converter(SubscriptionStatus)
    )
    created: str = attr.field(converter=str_to_datetime)
    updated: str = attr.field(converter=str_to_datetime)
    next_renewal_date: Optional[str] = attr.field(
        default=None, converter=str_to_datetime
    )
    previous_renewal_date: Optional[str] = attr.field(
        default=None, converter=str_to_datetime
    )

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the object.

        :return:
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    @property
    def overview(self) -> str:
        """
        Return a string that represents an overview of the model.

        :return:
        """
        deidentified = deepcopy(self)
        deidentified.deidentify()
        overview_dict = deidentified.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)

    def deidentify(self) -> None:
        """
        Remove unique database IDs from the base object.
        """
        deidentify(self)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict, Optional, Type, TypeVar, cast

from attr import fields, has
from omegaconf import OmegaConf
from omegaconf.errors import ConfigKeyError, ConfigTypeError, MissingMandatoryValue

OutputTypeVar = TypeVar("OutputTypeVar")


def deserialize_dictionary(
    input_dictionary: Dict[str, Any], output_type: Type[OutputTypeVar]
) -> OutputTypeVar:
    """
    Deserialize an `input_dictionary` to an object of the type passed in `output_type`.

    :param input_dictionary: Dictionary to deserialize
    :param output_type: Type of the object that the dictionary represents, and to
        which the data will be deserialized
    :return: Object of type `output_type`, holding the data passed in
        `input_dictionary`.
    """

    def prune_dict(data: dict, cls: Type[Any]) -> dict:
        """Recursively prune a dictionary to match the structure of an attr class."""
        pruned_data = {}
        for attribute in fields(cls):
            key = attribute.name
            if key in data:
                value = data[key]
                # Check if the field is itself a structured class
                if has(attribute.type) and isinstance(value, dict):
                    # Recursively prune the nested dictionary
                    pruned_data[key] = prune_dict(value, attribute.type)
                else:
                    pruned_data[key] = value
        return pruned_data

    filtered_input_dictionary = prune_dict(input_dictionary, output_type)
    model_dict_config = OmegaConf.create(filtered_input_dictionary)
    schema = OmegaConf.structured(output_type)
    schema_error: Optional[DataModelMismatchException] = None
    try:
        values = OmegaConf.merge(schema, model_dict_config)
        output = cast(output_type, OmegaConf.to_object(values))
    except (ConfigKeyError, MissingMandatoryValue, ConfigTypeError) as error:
        schema_error = DataModelMismatchException(
            input_dictionary=filtered_input_dictionary,
            output_data_model=output_type,
            message=error.args[0],
            error_type=type(error),
        )
        output = None
    if schema_error is not None:
        raise schema_error
    return output


class DataModelMismatchException(BaseException):
    """
    Exception raised when a deserialization event fails, meaning that the
    serialized input data does not match the expected schema for the object to
    construct.

    :param output_data_model: Type of the data model to which the data would be
        deserialized
    :param input_dictionary: Input dictionary which would be deserialized
    :param message: Error message that describes the error that occurred during
        deserialization
    :param error_type: Type of the error which occurred during deserialization
    """

    def __init__(
        self,
        input_dictionary: dict,
        output_data_model: Type,
        message: str,
        error_type: Type,
    ) -> None:
        self.output_data_model = output_data_model
        self.message = message
        self.input_dictionary = input_dictionary
        self.error_type = error_type

    def __str__(self) -> str:
        """
        Return the string representation of the exception.

        :return: String representation of the exception
        """
        return (
            f"Deserialization of input dictionary to object of type "
            f"'{self.output_data_model.__name__}' failed with error: \n\n"
            f"'{self.error_type.__name__}': {self.message}. "
            f"\n\nThe following input data was received for "
            f"deserialization: \n{self.input_dictionary}"
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .constants import BASE_TEST_PATH
from .enums import SdkTestMode
from .finalizers import force_delete_project
from .fixtures import get_sdk_fixtures
from .plotting import plot_predictions_side_by_side
from .project_helpers import get_or_create_annotated_project_for_test_class
from .project_service import ProjectService
from .training import attempt_to_train_task, await_training_start
from .vcr_helpers import are_cassettes_available, replace_unique_entries_in_cassettes

__all__ = [
    "BASE_TEST_PATH",
    "SdkTestMode",
    "get_sdk_fixtures",
    "ProjectService",
    "are_cassettes_available",
    "replace_unique_entries_in_cassettes",
    "get_or_create_annotated_project_for_test_class",
    "force_delete_project",
    "plot_predictions_side_by_side",
    "attempt_to_train_task",
    "await_training_start",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import glob
import logging
import os
import shutil
import time
from typing import Tuple

from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from .constants import CASSETTE_EXTENSION


def are_cassettes_available(cassette_path) -> bool:
    """
    Checks that the VCR cassettes required to run the tests offline are available

    :return: True if the cassettes are available in the proper path, False otherwise
    """
    if not os.path.isdir(cassette_path):
        return False
    if len(os.listdir(cassette_path)) > 0:
        return True
    return False


def replace_unique_entries_in_cassettes(
    entry_pairs: Tuple[Tuple[str, str]], cassette_dir: str
) -> None:
    """
    This function searches for the unique_entry in a target cassette file and
    replaces all occurrences of that address by a dummy value. The cassette file is
    updated in place

    :param entry_pairs: (unique_entry, dummy_value) pairs to search for and replace
    :param cassette_dir: Path to the directory containing the cassette files to search in
    """
    logging.info(
        f"Removing tre following entries from test cassettes: '{(entry_pair[0] for entry_pair in entry_pairs)}' ..."
    )
    tqdm_prefix = "Scrubbing cassettes"

    t_start = time.time()
    cassette_paths = glob.glob(os.path.join(cassette_dir, f"*.{CASSETTE_EXTENSION}"))

    with logging_redirect_tqdm(tqdm_class=tqdm):
        for cassette_path in tqdm(cassette_paths, desc=tqdm_prefix):
            replace_unique_entries_in_cassette(
                entry_pairs=entry_pairs, path_to_cassette_file=cassette_path
            )
    logging.info(
        f"Entries scrubbed from {len(cassette_paths)} cassette files in "
        f"{1000*(time.time()-t_start)} seconds"
    )


def replace_unique_entries_in_cassette(
    entry_pairs: Tuple[Tuple[str, str]], path_to_cassette_file: str
) -> None:
    """
    This function searches for the unique_entry in a target cassette file and
    replaces all occurrences of that address by a dummy value. The cassette file is
    updated in place

    :param entry_pairs: (unique_entry, dummy_value) pairs to search for and replace
    :param path_to_cassette_file: Path to the cassette file to search in
    """
    path_to_scrubbed_cassette_file = path_to_cassette_file + "_new"

    with open(path_to_cassette_file, "rt") as read_file, open(
        path_to_scrubbed_cassette_file, "wt"
    ) as write_file:
        for line in read_file.readlines():
            for unique_entry, dummy_value in entry_pairs:
                line = line.replace(unique_entry, dummy_value)
            write_file.write(line)

    os.remove(path_to_cassette_file)
    shutil.move(src=path_to_scrubbed_cassette_file, dst=path_to_cassette_file)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceTrigger


class AlwaysTrigger(PostInferenceTrigger):
    """
    Post inference trigger that activates on each inference call.
    """

    def __init__(self):
        super().__init__()
        self._repr_info_ = ""

    def __call__(self, image: np.ndarray, prediction: Prediction) -> float:
        """
        Compute a trigger score for the `image` and corresponding `prediction`.

        :param image: Numpy array representing an image
        :param prediction: Prediction object corresponding to the inference result
            for the image.
        :return: Float representing the score for the input
        """
        return 1



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
"""Utility functions for the post inference hook module"""
import time
from threading import Lock

try:
    from collections.abc import Iterator
except ImportError:
    # In python < 3.10 Iterator is in 'collections'
    from collections import Iterator


class RateLimiter(Iterator):
    """
    Iterator that yields at a maximum frequency of `frames_per_second`.

    :param frames_per_second: Maximum execution rate, defined in frames per second
    :param is_blocking: Setting this to True will block execution until the next
        yield. Setting this to False will cause the Iterator to yield `False` if it
        is called before the next yield is available, and True otherwise
    """

    def __init__(self, frames_per_second: float = 1, is_blocking: bool = False) -> None:
        self.interval = 1 / frames_per_second
        self.is_blocking = is_blocking
        self.__next_yield = 0
        self.__lock = Lock()

    def __iter__(self):
        """
        Return the iterator
        """
        return self

    def __next__(self) -> bool:
        """
        Return `True` if called after the `self.interval` has elapsed with respect to
        the last call. If `self.is_blocking == True`, will block execution otherwise,
        until the interval has elapsed. If `self.is_blocking == False`, execution is
        not blocked but `False` is returned instead.
        """
        with self.__lock:
            t = time.monotonic()
            if t < self.__next_yield:
                if self.is_blocking:
                    time.sleep(self.__next_yield - t)
                    t = time.monotonic()
                else:
                    return False
            self.__next_yield = t + self.interval
            return True



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Optional

import attr

from geti_sdk.data_models.enums import AnnotationState
from geti_sdk.data_models.utils import str_to_enum_converter_by_name_or_value


@attr.define
class TaskAnnotationState:
    """
    Representation of the state of an annotation for a particular task in an
    Intel® Geti™ project.
    """

    task_id: str
    state: Optional[str] = attr.field(
        converter=str_to_enum_converter_by_name_or_value(
            AnnotationState, allow_none=True
        ),
        default=None,
    )



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import csv
import itertools
import logging
import os
import time
from typing import Any, Dict, List, Optional, Sequence, Union

import cv2
import numpy as np
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk import Geti
from geti_sdk.data_models import (
    Image,
    Model,
    OptimizedModel,
    Performance,
    Project,
    Video,
)
from geti_sdk.deployment import Deployment
from geti_sdk.prediction_visualization.visualizer import Visualizer
from geti_sdk.rest_clients import ImageClient, ModelClient, TrainingClient, VideoClient
from geti_sdk.rest_clients.prediction_client import PredictionClient
from geti_sdk.utils.plot_helpers import (
    concat_prediction_results,
    pad_image_and_put_caption,
)

from .utils import get_system_info, load_benchmark_media, suppress_log_output


class Benchmarker:
    """
    Initialize and manage benchmarking experiments to measure model throughput on
    different hardware.
    """

    def __init__(
        self,
        geti: Geti,
        project: Project,
        precision_levels: Optional[Sequence[str]] = None,
        models: Optional[Sequence[Model]] = None,
        algorithms: Optional[Sequence[str]] = None,
        benchmark_images: Optional[
            Sequence[Union[Image, np.ndarray, os.PathLike]]
        ] = None,
        benchmark_video: Optional[Union[Video, os.PathLike]] = None,
    ):
        """
        Manage benchmarking experiments to measure inference model throughput on
        different hardware. It allows for quick and easy comparison of inference
        framerates for different model architectures and precision levels for the
        specified project.

        The Benchmarker will fetch models from Intel Geti and measure the
        throughput for local inference for these models. It requires an existing Geti
        Project to work. The project does not need to be trained yet, but must have
        sufficient annotations to be able to start training.

        The benchmark can be run on any target device that is supported by OpenVINO.
        Specifying multiple algorithms and precision levels to include in the
        benchmark allows for easy comparison of the performance for these models.

        NOTE: For task chain projects, passing `models` or `algorithms` directly in the
        constructor is not supported. The Benchmarker provides dedicated
        methods to specify which models or algorithms to use in this case, which can
        be called after initialization.

        :param geti: Geti instance on which the project to use for benchmarking lives
        :param project: Project to use for the benchmarking. The
            project must exist on the specified Geti instance
        :param precision_levels: List of model precision levels to run the
            benchmarking for. Throughput will be measured for each precision level
            in this list. Valid options are ["FP32", "FP16", "INT8"]. Defaults to:
            ["FP32", "FP16"]
        :param models: Optional list of base models to use in the benchmarking.
            Note that this should not be optimized models: The Benchmarker will fetch
            optimized models in the specified `precision_levels` for the base
            models specified here
        :param algorithms: List of algorithm names for which the benchmarking should
            be performed. If no models are trained for all algorithms in the list,
            the Benchmarker will trigger training for those algorithms and wait for
            the training jobs to complete. If neither `models` or `algorithms` are
            specified, the Benchmarker uses the current active model(s) in the project.
        :param benchmark_images: Optional list of images to run the benchmarks on. Can
            be passed as Geti Images, numpy images or filepaths to image files on disk.
            NOTE: If no images are specified, the benchmark will use a random image
            or video frame from the project
        :param benchmark_video: Optional video to use for the benchmarking. Can be
            passed either as a Geti Video object, or as a filepath to a video file on
            disk.
            NOTE: Either `benchmark_video` or `benchmark_images` can be specified, but
            not both. If neither images nor video is passed, the Benchmarker will
            pick a random image or video frame from the project to run the benchmark
            on.
        """
        self.geti = geti
        # Update project object to get the latest project details
        self.project = self.geti.get_project(project_id=project.id)
        logging.info(
            f"Setting up Benchmarker for Intel® Geti™ project `{self.project.name}`."
        )
        self._is_single_task = len(self.project.get_trainable_tasks()) == 1
        if precision_levels is None:
            precision_levels = ["FP32", "FP16"]
        self.precision_levels = precision_levels
        self.model_client = ModelClient(
            session=geti.session, workspace_id=geti.workspace_id, project=self.project
        )
        self.training_client = TrainingClient(
            session=geti.session, workspace_id=geti.workspace_id, project=self.project
        )

        self._models: Optional[List[Model]] = None
        self._task_chain_models: Optional[List[List[Model]]] = None
        self._algorithms: Optional[List[str]] = None
        self._task_chain_algorithms: Optional[List[List[str]]] = None
        self._optimized_models: Optional[List[OptimizedModel]] = None
        self._task_chain_optimized_models: Optional[List[List[OptimizedModel]]] = None
        self._deployment_folders: List[os.PathLike] = []

        if not self._is_single_task and (models is not None or algorithms is not None):
            raise ValueError(
                "You have specified a task-chain project to be used for benchmarking. "
                "This does not allow setting `models` or `algorithms` via the "
                "Benchmark constructor. Please specify which models or algorithms "
                "should be used in the benchmark using the methods "
                "`Benchmarker.set_task_chain_models()` or "
                "`Benchmarker.set_task_chain_algorithms()`. These methods "
                "can be used after initialization of the Benchmarker. "
                "Please refer to the method documentation for further guidance."
            )

        if models is None and algorithms is None:
            if self._is_single_task:
                logging.info(
                    "No `models` or `algorithms` were specified, using current active "
                    "models in the project."
                )
            else:
                logging.info(
                    "Models or algorithms to benchmark for a task chain project can "
                    "be specified by calling the `Benchmarker.set_task_chain_models()`"
                    " or `Benchmarker.set_task_chain_algorithms()` methods. If these "
                    "methods are not called, the benchmark will be set up using the "
                    "current active models."
                )
            models = self.model_client.get_all_active_models()
            if len(models) == 0:
                raise ValueError(
                    "No trained models were found in the project, please either "
                    "train a model first or specify an algorithm to train."
                )
            algorithms: List[str] = []
            for model in models:
                task = self.model_client.get_task_for_model(model=model)
                if self._is_single_task:
                    logging.info(
                        f"Found active model `{model.name}` for task `{task.title}`"
                    )
                algorithms.append(model.architecture)
            if self._is_single_task:
                self._models = models
                self._algorithms = algorithms
            else:
                self._task_chain_models = [[model] for model in models]
                self._task_chain_algorithms = [[algo] for algo in algorithms]
            self._are_models_specified = True

        elif models is not None and algorithms is not None:
            raise ValueError(
                "Either `models` or `algorithms` could be specified, but not both."
            )
        elif models is not None and algorithms is None:
            self._models = models
            self._are_models_specified = True
            self._algorithms = [model.architecture for model in models]
        elif models is None and algorithms is not None:
            self._models = None
            self._are_models_specified = False
            self._algorithms = algorithms

        self.images = None
        self.video = None
        if benchmark_video is not None and benchmark_images is not None:
            raise ValueError(
                "Please specify either `benchmark_video` or `benchmark_images`, but "
                "not both."
            )
        elif benchmark_images is not None:
            self.images = benchmark_images
        elif benchmark_video is not None:
            self.video = benchmark_video
        else:
            image_client = ImageClient(
                session=geti.session,
                workspace_id=geti.workspace_id,
                project=self.project,
            )
            images = image_client.get_all_images()
            if len(images) > 0:
                self.images = [images[0]]
            else:
                video_client = VideoClient(
                    session=geti.session,
                    workspace_id=geti.workspace_id,
                    project=self.project,
                )
                videos = video_client.get_all_videos()
                if len(videos) > 0:
                    self.video = videos[0]
                else:
                    raise ValueError(
                        "No benchmark images or video was specified, and no media "
                        "could be found in the project. Unable to proceed, please "
                        "specify either images or a video to run the benchmark on."
                    )

    def set_task_chain_models(
        self, models_task_1: Sequence[Model], models_task_2: Sequence[Model]
    ):
        """
        Set the models to be used in the benchmark for a task-chain project. The
        benchmarking will run for all possible combinations of models for task 1
        and task 2.

        :param models_task_1: Models to use for task #1
        :param models_task_2: Models to use for task #2
        """
        if self._is_single_task:
            logging.warning(
                "Method `set_task_chain_models` was called for a non-task-chain "
                "project. This has no effect."
            )
            return
        self._task_chain_models = [
            list(pair) for pair in itertools.product(models_task_1, models_task_2)
        ]
        self._are_models_specified = True
        self._task_chain_algorithms = [
            [model.architecture for model in pair] for pair in self._task_chain_models
        ]
        logging.info(
            f"Task chain models set. Found a total of "
            f"{len(self._task_chain_models)} possible combinations to benchmark"
        )

    def set_task_chain_algorithms(
        self, algorithms_task_1: Sequence[str], algorithms_task_2: Sequence[str]
    ):
        """
        Set the algorithms to be used in the benchmark for a task-chain project. The
        benchmarking will run for all possible combinations of algorithms for task 1
        and task 2.

        Note that upon benchmark initialization, the Benchmarker will check if
        trained models are available for all algorithms specified

        :param algorithms_task_1: Algorithms to use for task #1
        :param algorithms_task_2: Algorithms to use for task #2
        """
        if self._is_single_task:
            logging.warning(
                "Method `set_task_chain_algorithms` was called for a non-task-chain "
                "project. This has no effect."
            )
            return
        self._task_chain_algorithms = [
            list(pair)
            for pair in itertools.product(algorithms_task_1, algorithms_task_2)
        ]
        self._are_models_specified = False
        self._task_chain_models = None
        logging.info(
            f"Task chain algorithms set. Found a total of "
            f"{len(self._task_chain_algorithms)} possible algorithm combinations "
            f"to benchmark"
        )

    @property
    def models(self) -> Union[List[Model], List[List[Model]]]:
        """
        Return the models to be used in the benchmark.
        """
        if self._are_models_specified:
            if self._is_single_task:
                return self._models
            return self._task_chain_models
        raise ValueError(
            "Unable to access models, no benchmark models have been specified yet."
        )

    @property
    def algorithms(self) -> Union[List[str], List[List[str]]]:
        """
        Return the algorithm names to be used in the benchmark
        """
        if self._is_single_task:
            return self._algorithms
        return self._task_chain_algorithms

    @property
    def optimized_models(
        self,
    ) -> Union[List[OptimizedModel], List[List[OptimizedModel]]]:
        """
        Return the optimized models to be used in deployments for the benchmark
        """
        if self._is_single_task and self._optimized_models is not None:
            return self._optimized_models
        elif not self._is_single_task and self._task_chain_optimized_models is not None:
            return self._task_chain_optimized_models
        raise ValueError(
            "Optimized models have not been assigned yet. Please initialize the "
            "benchmarker first using the `initialize_benchmark` method."
        )

    def _train_model_for_algorithm(self, task_index: int, algorithm_name: str) -> Model:
        """
        Train a model for the specified task and algorithm

        :param task_index: Index of the task for which to train the model
        :param algorithm_name: Name of the algorithm to use
        :return: Model object, representing the trained model
        """
        algos = self.training_client.get_algorithms_for_task(task_index)
        algo = algos.get_by_name(algorithm_name)
        job = self.training_client.train_task(task_index, algorithm=algo)
        job = self.training_client.monitor_job(job=job)
        return self.model_client.get_model_for_job(job)

    def _optimize_model_for_algorithm(
        self, model: Model, precision: str = "INT8"
    ) -> OptimizedModel:
        """
        Optimize a `model` with the given `precision`.

        :param model: Base model to optimize
        :param precision: Precision to use for the optimization
        :return: OptimizedModel object, representing the optimized model
        """
        optimization_type_map = {"INT8": "pot"}
        job = self.model_client.optimize_model(
            model=model, optimization_type=optimization_type_map[precision]
        )
        self.model_client.monitor_job(job)
        updated_model = self.model_client.update_model_detail(model)
        return updated_model.get_optimized_model(
            precision=precision, optimization_type="openvino"
        )

    def prepare_benchmark(self, working_directory: os.PathLike = "."):
        """
        Prepare the benchmarking experiments. This involves:

            1. Ensuring that all required models are available, i.e. that all
                specified algorithms have a trained model in the Geti project. If
                not, training jobs will be started and awaited.
            2. Ensuring that for each model, optimized models with the required
                quantization level are available. If not, optimization jobs will
                be started and awaited.
            3. Creating and downloading deployments for all models to benchmark.

        :param working_directory: Output directory to which the deployments for the
            benchmark will be saved.
        """
        logging.info("Preparing benchmark experiments.")
        # First, check if model training is required. This happens when
        # algorithms are specified, but not all algorithms may have a model
        # trained already.
        if not self._are_models_specified:
            if self._is_single_task:
                logging.info(
                    f"Checking model availability for {len(self.algorithms)} "
                    f"different algorithms: {self.algorithms}."
                )
                models: List[Model] = []
                for algorithm_name in self.algorithms:
                    model = self.model_client.get_latest_model_by_algo_name(
                        algorithm_name
                    )
                    if model is None:
                        logging.info(
                            f"No model found in project for algorithm "
                            f"{algorithm_name}, requesting model training"
                        )
                        model = self._train_model_for_algorithm(
                            task_index=0, algorithm_name=algorithm_name
                        )
                    models.append(model)
                self._models = models
            else:
                logging.info(
                    f"Checking model availability for {len(self.algorithms)} "
                    f"different pairs of algorithms."
                )
                models: List[List[Model]] = []
                for algo_pair in self.algorithms:
                    model_pair: List[Model] = []
                    for task_index, algorithm_name in enumerate(algo_pair):
                        model = self.model_client.get_latest_model_by_algo_name(
                            algorithm_name
                        )
                        if model is None:
                            model = self._train_model_for_algorithm(
                                task_index=task_index, algorithm_name=algorithm_name
                            )
                        model_pair.append(model)
                    models.append(model_pair)
                self._task_chain_models = models
            self._are_models_specified = True
        logging.info("All required base models are available")
        # Then, check if model optimization is required. Create a list of
        # optimized models for which deployments should be created.
        if self._is_single_task:
            logging.info(
                f"Checking optimized model availability for {len(self.models)} "
                f"different models and quantization levels: {self.precision_levels}"
            )
            optimized_models: List[OptimizedModel] = []
            for model in self.models:
                for precision in self.precision_levels:
                    opt_model = model.get_optimized_model(
                        precision=precision, optimization_type="openvino"
                    )
                    if opt_model is None:
                        logging.info(
                            f"No optimized model with quantization level {precision} "
                            f"found for model {model.name}, requesting model "
                            f"optimization."
                        )
                        opt_model = self._optimize_model_for_algorithm(
                            model=model, precision=precision
                        )
                    optimized_models.append(opt_model)
            self._optimized_models = optimized_models
        else:
            logging.info(
                f"Checking optimized model availability for {len(self.models)} "
                f"different model pairs and quantization levels: "
                f"{self.precision_levels}"
            )
            optimized_models: List[List[OptimizedModel]] = []
            for model_pair in self.models:
                for precision in self.precision_levels:
                    opt_model_pair: List[OptimizedModel] = []
                    for model in model_pair:
                        # Update model info, to take into account any
                        # optimizations made during loop execution
                        model = self.model_client.update_model_detail(model=model)
                        opt_model = model.get_optimized_model(
                            precision=precision, optimization_type="openvino"
                        )
                        if opt_model is None:
                            logging.info(
                                f"No optimized model with quantization level {precision} "
                                f"found for model {model.name}, requesting model "
                                f"optimization."
                            )
                            opt_model = self._optimize_model_for_algorithm(
                                model=model, precision=precision
                            )
                        opt_model_pair.append(opt_model)
                    optimized_models.append(opt_model_pair)
            self._task_chain_optimized_models = optimized_models
        logging.info("All required optimized models are available")

        # Next up, create deployments for all optimized models
        os.makedirs(working_directory, exist_ok=True)
        logging.info(
            f"Creating {len(self.optimized_models)} deployments to benchmark. Saving "
            f"deployment data to folder: `{working_directory}`"
        )
        with logging_redirect_tqdm(tqdm_class=tqdm):
            for index, opt_models in tqdm(
                enumerate(self.optimized_models),
                total=len(self.optimized_models),
                desc="Creating deployments",
            ):
                if isinstance(opt_models, OptimizedModel):
                    opt_models = [opt_models]
                output_folder = os.path.join(working_directory, f"deployment_{index}")
                with suppress_log_output():
                    self.geti.deploy_project(
                        project=self.project,
                        output_folder=output_folder,
                        models=opt_models,
                    )
                    self._deployment_folders.append(output_folder)
        logging.info("Deployments created. Benchmark initialization complete.")

    def initialize_from_folder(self, target_folder: os.PathLike = "."):
        """
        Initialize the Benchmarker from a folder containing deployments. This method
        checks that any directory inside the `target_folder` contains a valid
        deployment for the project assigned to this Benchmarker.

        :param target_folder: Directory containing the model deployments that should
            be used in the Benchmarking.
        """
        logging.info(f"Initializing Benchmarker from folder `{target_folder}`")
        if len(self._deployment_folders) != 0:
            raise ValueError(
                "The Benchmarker appears to be already initialized. Unable to "
                "continue."
            )
        for object in os.listdir(target_folder):
            object_path = os.path.join(target_folder, object)
            if os.path.isfile(object_path):
                continue
            try:
                deployment = Deployment.from_folder(path_to_folder=object_path)
            except ValueError:
                logging.info(
                    f"Directory {object_path} does not contain a valid Deployment, "
                    f"skipping..."
                )
                continue
            if deployment.project.name != self.project.name:
                raise ValueError(
                    f"Deployment at path `{object_path}` was created for project "
                    f"{deployment.project.name}, yet the Benchmarker was set up for "
                    f"project {self.project.name}. Unable to initialize. Please use "
                    f"only deployments created for project `{self.project.name}`"
                )
            self._deployment_folders.append(object_path)

        if len(self._deployment_folders) != 0:
            logging.info(
                f"Benchmarker was initialized from folder. {len(self._deployment_folders)} "
                f"valid deployment folders were found."
            )

    def run_throughput_benchmark(
        self,
        working_directory: os.PathLike = ".",
        results_filename: str = "results",
        target_device: str = "CPU",
        frames: int = 200,
        repeats: int = 3,
    ) -> List[Dict[str, str]]:
        """
        Run the benchmark experiment.

        :param working_directory: Directory in which the deployments that should be
            benchmarked are stored. All output will be saved to this directory.
        :param results_filename: Name of the file to which the results will be saved.
            File extension should not be included, the results will always be saved as
            a `.csv` file. Defaults to `results.csv`. The results file will be created
            within the `working_directory`
        :param target_device: Device to run the inference models on, for example "CPU"
            or "GPU". Defaults to "CPU".
        :param frames: Number of frames/images to infer in order to calculate
            fps
        :param repeats: Number of times to repeat the benchmark runs. FPS will be
            averaged over the runs.
        """
        if len(self._deployment_folders) == 0:
            raise ValueError(
                "Benchmarker does not contain any deployments to benchmark yet! Please "
                "prepare the deployments first using either the "
                "`Benchmarker.prepare_benchmark()` or "
                "`Benchmarker.initialize_from_folder()` methods."
            )
        logging.info("Starting throughput benchmark experiments.")

        logging.info(
            f"The Benchmarker will run for {len(self._deployment_folders)} deployments"
        )

        logging.info("Loading benchmark media")
        benchmark_frames = load_benchmark_media(
            session=self.geti.session,
            images=self.images,
            video=self.video,
            frames=frames,
        )

        results_file = os.path.join(working_directory, f"{results_filename}.csv")
        logging.info(f"Writing results to `{results_file}`")

        logging.info(
            f"Benchmarking inference rate for synchronous inference on {frames} frames "
            f"with {repeats} repeats"
        )
        with logging_redirect_tqdm(tqdm_class=tqdm):
            results: List[Dict[str, str]] = []
            for deployment_index, deployment_folder in enumerate(
                tqdm(self._deployment_folders, desc="Benchmarking")
            ):
                success = True
                deployment = Deployment.from_folder(deployment_folder)
                try:
                    with suppress_log_output():
                        deployment.load_inference_models(device=target_device)
                except Exception as e:
                    success = False
                    logging.info(
                        f"Failed to load inference models for deployment at path: "
                        f"`{deployment_folder}`, with error: {e}. Marking benchmark "
                        f"run for the deployment as failed"
                    )

                if success:
                    try:
                        # Warm-up for the model
                        deployment.infer(benchmark_frames[0])

                        # Estimate time to completion
                        t_single_start = time.time()
                        deployment.infer(benchmark_frames[0])
                        single_inf_time = time.time() - t_single_start
                        logging.info(
                            f"Inference model(s) for deployment `{deployment_folder}` "
                            f"loaded. Starting benchmark run. Estimated time required: "
                            f"{repeats*frames*single_inf_time:.0f} seconds"
                        )
                    except Exception as e:
                        success = False
                        logging.info(
                            f"Failed to run inference on frame number 0. Marking "
                            f"benchmark run for deployment `{deployment_folder}` as "
                            f"failed. Inference failed with error: `{e}`"
                        )
                    t_start = time.time()
                    for run in range(repeats):
                        for frame in benchmark_frames:
                            try:
                                deployment.infer(frame)
                            except Exception as e:
                                success = False
                                logging.info(
                                    f"Failed to run inference on frame number "
                                    f"{benchmark_frames.index(frame)}. Marking "
                                    f"benchmark run for deployment "
                                    f"`{deployment_folder}` as failed. Inference "
                                    f"failed with error: `{e}`"
                                )
                    t_elapsed = time.time() - t_start
                    total_frames = frames * repeats
                    fps = t_elapsed and total_frames / t_elapsed or 0
                else:
                    fps = 0

                model_scores = []
                for om in deployment.models:
                    if isinstance(om.performance, Performance):
                        score = om.performance.score
                    elif isinstance(om.performance, dict):
                        score = om.performance.get("score", -1)
                    else:
                        score = -1
                    model_scores.append(score)

                # Update result list
                result_row: Dict[str, str] = {}
                result_row["name"] = f"Deployment {deployment_index}"
                result_row["project_name"] = self.project.name
                result_row["target_device"] = target_device
                result_row["task 1"] = self.project.get_trainable_tasks()[0].title
                result_row["model 1"] = deployment.models[0].name
                result_row["model 1 score"] = f"{model_scores[0]:.2f}"
                if not self._is_single_task:
                    result_row["task 2"] = self.project.get_trainable_tasks()[1].title
                    result_row["model 2"] = deployment.models[1].name
                    result_row["model 2 score"] = f"{model_scores[1]:.2f}"
                result_row["success"] = str(int(success))
                result_row["fps"] = f"{fps:.2f}"
                result_row["total frames"] = f"{frames * repeats}"
                result_row["source"] = deployment_folder
                result_row.update(get_system_info(device=target_device))

                # Write results to file
                fieldnames = list(result_row.keys())
                if not results:  # First row
                    with open(results_file, "w", newline="") as csvfile:
                        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                        writer.writeheader()
                        writer.writerow(result_row)
                else:  # Append
                    with open(results_file, "a", newline="") as csvfile:
                        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                        writer.writerow(result_row)
                results.append(result_row)

        return results

    def _predict_using_active_model(
        self,
        numpy_image: np.ndarray,
    ) -> Dict[str, Any]:
        """
        Predict on platform using the active model.

        :param numpy_image: Numpy array containing the image to be predicted on.
        :return: Dictionary containing the prediction results.
        """
        # Predict on the Platform
        prediction_client = PredictionClient(
            session=self.geti.session,
            workspace_id=self.geti.workspace_id,
            project=self.project,
        )
        platform_prediction = prediction_client.predict_image(numpy_image)

        # load active models info
        active_models = self.model_client.get_all_active_models()
        result: dict[str, Any] = {}
        result["prediction"] = platform_prediction
        result["run_name"] = "On-Platform Prediction"
        result["model_1"] = active_models[0].name + " " + active_models[0].precision[0]
        result["model_1_score"] = active_models[0].performance.score
        if not self._is_single_task:
            result["model_2"] = (
                active_models[1].name + " " + active_models[1].precision[0]
            )
            result["model_2_score"] = active_models[1].performance.score
        return result

    def _add_header_to_comparison(
        self, comparison_image: np.ndarray, target_device: str
    ) -> np.ndarray:
        """
        Add a header to the comparison image.

        :param comparison_image: Comparison image to add the header to.
        :return: Comparison image with header.
        """
        # Calculate text and image padding size
        text_scale = round(comparison_image.shape[1] / 1280, 1)
        thickness = int(text_scale / 1.4)
        (_, label_height), baseline = cv2.getTextSize(
            "Test caption", cv2.FONT_HERSHEY_SIMPLEX, text_scale, thickness
        )
        top_padding_per_line = label_height + baseline
        # Prepare image captions
        device_info = get_system_info(device=target_device)["device_info"]
        caption_lines = [
            "Inference results comparison",
            f"Project: {self.project.name}",
            (
                f"Task: {self.project.get_trainable_tasks()[0].title}"
                + (
                    ""
                    if self._is_single_task
                    else f" -> {self.project.get_trainable_tasks()[1].title}"
                )
            ),
            f"Device info: {device_info}",
        ]
        # Pad the image
        padded_image = cv2.copyMakeBorder(
            comparison_image,
            top=2 * baseline + top_padding_per_line * len(caption_lines),
            bottom=0,
            left=0,
            right=0,
            borderType=cv2.BORDER_CONSTANT,
            value=(255, 255, 255),
        )
        # Put text
        for line_number, text_line in enumerate(caption_lines):
            cv2.putText(
                padded_image,
                text_line,
                (10, top_padding_per_line * (line_number + 1)),
                cv2.FONT_HERSHEY_SIMPLEX,
                text_scale,
                (0, 0, 0),
                thickness,
            )
        return padded_image

    def compare_predictions(
        self,
        working_directory: os.PathLike = ".",
        saved_image_name: str = "comparison",
        target_device: str = "CPU",
        image: Optional[Union[np.ndarray, str, os.PathLike]] = None,
        include_online_prediction_for_active_model: bool = True,  # the name is not finalized
        throughput_benchmark_results: Optional[
            Union[List[Dict[str, str]], os.PathLike]
        ] = None,
    ) -> np.ndarray:
        """
        TODO blank image if not success

        Perform visual comparison of predictions from different deployments.

        :param working_directory: Directory in which the deployments that should be
            benchmarked are stored. All output will be saved to this directory.
        :param saved_image_name: Name of the file to which the results will be saved.
            File extension should not be included, the results will always be saved as
            a `.jpg` file. Defaults to `comparison.jpg`. The results file will be created
            within the `working_directory`
        :param target_device: Device to run the inference models on, for example "CPU"
            or "GPU". Defaults to "CPU".
        :param image: Image to use for comparison. If no image is passed, the first
            image in the `images` list will be used.
        :param include_online_prediction_for_active_model: Flag to include prediction
            from the active model on the platform side.
        :param throughput_benchmark_results: Results from a throughput benchmark run. If
            this is passed, the captions for the images will contain the benchmark results.
        :return: Image containing visual comparison in form of a NumPy array.
        """
        if len(self._deployment_folders) == 0:
            raise ValueError(
                "Benchmarker does not contain any deployments to benchmark yet! Please "
                "prepare the deployments first using either the "
                "`Benchmarker.prepare_benchmark()` or "
                "`Benchmarker.initialize_from_folder()` methods."
            )
        logging.info("Starting collecting predictions for visual comparison.")

        logging.info(
            f"The Benchmarker will run for {len(self._deployment_folders)} deployments"
        )

        logging.info("Loading benchmark media")
        if isinstance(image, np.ndarray):
            pass
        elif image is None:
            image = load_benchmark_media(
                session=self.geti.session,
                images=self.images,
                video=self.video,
                frames=1,
            )[0]
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        elif isinstance(image, (str, os.PathLike)):
            image = cv2.cvtColor(cv2.imread(image), cv2.COLOR_BGR2RGB)
        else:
            raise TypeError(f"Invalid image type: {type(image)}.")

        saved_image_path = os.path.join(working_directory, f"{saved_image_name}.jpg")
        logging.info(f"Saving visual comparison to `{saved_image_path}`")

        # Check the benchmark results
        if isinstance(throughput_benchmark_results, (os.PathLike, str)):
            with open(throughput_benchmark_results, "r") as results_file:
                throughput_benchmark_results = list(csv.DictReader(results_file))

        visualizer = Visualizer()

        # Performe inferece
        with logging_redirect_tqdm(tqdm_class=tqdm):
            results: List[List[np.ndarray]] = []
            model_name_to_row: dict[str, int] = {}
            for deployment_index, deployment_folder in enumerate(
                tqdm(self._deployment_folders, desc="Collecting predictions")
            ):
                success = True
                deployment = Deployment.from_folder(deployment_folder)
                try:
                    with suppress_log_output():
                        deployment.load_inference_models(device=target_device)
                except Exception as e:
                    success = False
                    logging.info(
                        f"Failed to load inference models for deployment at path: "
                        f"`{deployment_folder}`, with error: {e}. Marking benchmark "
                        f"run for the deployment as failed"
                    )

                if success:
                    try:
                        prediction = deployment.infer(image)
                    except Exception as e:
                        success = False
                        logging.info(
                            f"Failed to run inference on the image. Marking "
                            f"benchmark run for deployment `{deployment_folder}` as "
                            f"failed. Inference failed with error: `{e}`"
                        )
                if success:
                    image_with_prediction = visualizer.draw(image, prediction)
                    image_with_prediction = cv2.cvtColor(
                        image_with_prediction, cv2.COLOR_BGR2RGB
                    )
                else:
                    # Replace the image with an empty one in case of no prediction
                    image_with_prediction = np.zeros_like(image)
                    image_with_prediction = cv2.putText(
                        image_with_prediction,
                        "Failed to run inference on the image",
                        (10, 50),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        1,
                        (255, 0, 0),
                        2,
                    )
                # Save model scores
                model_scores = []
                for om in deployment.models:
                    if isinstance(om.performance, Performance):
                        score = om.performance.score
                    elif isinstance(om.performance, dict):
                        score = om.performance.get("score", -1)
                    else:
                        score = -1
                    model_scores.append(score)

                model_info = {
                    "run_name": f"Deployment {deployment_index}",
                    "model_1": deployment.models[0].name,
                    "model_1_score": model_scores[0],
                    "fps": (
                        None
                        if throughput_benchmark_results is None
                        else throughput_benchmark_results[deployment_index]["fps"]
                    ),
                }
                if not self._is_single_task:
                    model_info.update(
                        {
                            "model_2": deployment.models[1].name,
                            "model_2_score": model_scores[1],
                        }
                    )
                # Pad the image and put captions on it
                image_with_prediction = pad_image_and_put_caption(
                    image=image_with_prediction, **model_info
                )

                # Image is ready, now we add it to the results array
                # Determine result's position on the grid
                model_1_name = deployment.models[0].name.split(" ")[0]
                if model_1_name not in model_name_to_row:
                    model_name_to_row[model_1_name] = len(results)
                    results.append([])
                row_n = model_name_to_row[model_1_name]
                results[row_n].append(image_with_prediction)

        if include_online_prediction_for_active_model:
            logging.info("Predicting on the platform using the active model")
            online_prediction_result = self._predict_using_active_model(image)
            image_with_prediction = visualizer.draw(
                image, online_prediction_result["prediction"]
            )
            image_with_prediction = cv2.cvtColor(
                image_with_prediction, cv2.COLOR_BGR2RGB
            )

            del online_prediction_result["prediction"]
            image_with_prediction = pad_image_and_put_caption(
                image=image_with_prediction, **online_prediction_result
            )
            # Add online prediction to a separate row
            results.append(
                [
                    image_with_prediction,
                ]
            )
        image_grid = concat_prediction_results(results=results)
        image_with_header = self._add_header_to_comparison(
            image_grid, target_device=target_device
        )

        # Save image to file
        cv2.imwrite(
            saved_image_path, cv2.cvtColor(image_with_header, cv2.COLOR_RGB2BGR)
        )
        return image_with_header



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import copy
import datetime
import logging
from concurrent.futures import ThreadPoolExecutor
from threading import BoundedSemaphore
from typing import Any, Dict, Optional

import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import (
    PostInferenceAction,
    PostInferenceHookInterface,
    PostInferenceTrigger,
)
from geti_sdk.post_inference_hooks.utils import RateLimiter


class PostInferenceHook(PostInferenceHookInterface):
    """
    Basic hook that can be executed after inference. A PostInferenceHook consists of:
      - A Trigger, defining a condition
      - An Action, defining an act or sequence of actions to perform when the
        trigger is activated.

    By default, hooks are executed in parallel in a separate thread from the main
    inference code. If you want to disable parallel execution, make sure to set
    `max_threads=0` when creating the hook.

    :param trigger: PostInferenceTrigger which is evaluated and used to trigger the
        hook action
    :param action: PostInferenceAction to be executed if triggered
    :param max_threads: Maximum number of threads to use for hook execution. Defaults
        to 5. Set to 0 to disable parallel hook execution.
    :param limit_action_rate: True to limit the rate at which the `action` for the
        hook is executed. Note that rate limiting is applied **after** trigger
        evaluation, so this can be useful for handling video streaming in which
        subsequent frames may be very similar. Enabling rate limiting will avoid
        sampling many near-duplicate frames
    :param max_frames_per_second: Maximum frame rate at which to execute the action.
        Only takes effect if `limit_action_rate = True`.
    :param queue_limit: Maximum size of the execution queue for the hook. If left as 0
        (the default), the queue is unbounded. This may lead to high memory usage in
        some cases (if inference is very fast, or the action execution is slow). When
        running into out of memory issues, try setting the queue limit to avoid
        buffering too many frames in memory. Note that this may slow down the
        inference rate
    """

    def __init__(
        self,
        trigger: PostInferenceTrigger,
        action: PostInferenceAction,
        max_threads: int = 5,
        limit_action_rate: bool = False,
        max_frames_per_second: float = 1,
        queue_limit: int = 0,
    ):
        super().__init__(trigger=trigger, action=action)

        self.parallel_execution = max_threads != 0
        self._semaphore: Optional[BoundedSemaphore] = None

        if self.parallel_execution:
            self.executor = ThreadPoolExecutor(max_workers=max_threads)
            logging.debug(
                f"Parallel inference hook execution enabled, using a maximum of "
                f"{max_threads} threads."
            )
            if queue_limit > 0:
                self._semaphore = BoundedSemaphore(queue_limit + max_threads)

        if limit_action_rate:
            self.rate_limiter: Optional[RateLimiter] = RateLimiter(
                frames_per_second=max_frames_per_second, is_blocking=False
            )
        else:
            self.rate_limiter = None

    def run(
        self,
        image: np.ndarray,
        prediction: Prediction,
        name: Optional[str] = None,
        timestamp: Optional[datetime.datetime] = None,
    ) -> None:
        """
        Run the post inference hook. First evaluate the trigger, then execute the
        action if the trigger conditions are met.

        :param image: Numpy array representing the image
        :param prediction: Prediction object containing the inference results for the
            image
        :param name: Optional name of the image which can be used in the hook
            action, for example as filename or tag for data collection
        :param timestamp: Optional timestamp belonging to the image
        """

        def execution_function(
            im: np.ndarray, pred: Prediction, ts: datetime.datetime
        ) -> None:
            score = self.trigger(image=im, prediction=pred)
            decision = self.trigger.get_decision(score=score)
            if decision:
                if self.rate_limiter is not None:
                    take_action = next(self.rate_limiter)
                else:
                    take_action = True
                if take_action:
                    self.action(
                        image=im, prediction=pred, score=score, name=name, timestamp=ts
                    )

        if timestamp is None:
            timestamp = datetime.datetime.now()
        if self.parallel_execution:
            if self._semaphore is not None:
                self._semaphore.acquire()
            try:
                future = self.executor.submit(
                    execution_function, image, prediction, timestamp
                )
            except Exception as error:
                if self._semaphore is not None:
                    self._semaphore.release()
                raise error
            if self._semaphore is not None:
                future.add_done_callback(lambda x: self._semaphore.release())
        else:
            execution_function(image, prediction, timestamp)

    def __del__(self):
        """
        Await shutdown of the ThreadPoolExecutor, if needed.

        Called upon deletion of the post inference hook or garbage collection
        """
        if self.parallel_execution:
            self.executor.shutdown(wait=True)

    def __repr__(self) -> str:
        """
        Return a string representation of the PostInferenceHook object

        :return: String representing the post inference hook
        """
        rate_msg = ""
        if self.rate_limiter is not None:
            rate_msg = (
                f"Action rate limited to {1/self.rate_limiter.interval:.1f} "
                f"frames per second."
            )
        thread_msg = ""
        if self.parallel_execution:
            thread_msg = "Multithreaded execution enabled."

        suffix_msg = f"[{thread_msg} {rate_msg}]"
        if len(suffix_msg) == 3:
            suffix_msg = ""
        return f"PostInferenceHook(trigger={self.trigger}, action={self.action}){suffix_msg}"

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "PostInferenceHook":
        """
        Create a PostInferenceHook object from it's dictionary representation

        :param input_dict: Dictionary representing the hook
        :return: Initialized PostInferenceHook object, with a trigger and action set
            according to the input dictionary
        """
        input_copy = copy.deepcopy(input_dict)

        # Construct trigger
        trigger = PostInferenceTrigger.from_dict(input_copy["trigger"])
        input_copy.update({"trigger": trigger})

        # Construct action
        action = PostInferenceAction.from_dict(input_copy["action"])
        input_copy.update({"action": action})

        return cls(**input_copy)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `demos` package contains useful functions for setting up demo projects on any
Intel® Geti™ server.

Module contents
---------------

"""
from .constants import DEFAULT_DATA_PATH, EXAMPLE_IMAGE_PATH
from .data_helpers import (
    get_coco_dataset,
    get_mvtec_dataset,
    get_person_car_bike_video,
    set_directory_permissions,
)
from .demo_projects import (
    create_anomaly_classification_demo_project,
    create_classification_demo_project,
    create_detection_demo_project,
    create_detection_to_classification_demo_project,
    create_detection_to_segmentation_demo_project,
    create_segmentation_demo_project,
    ensure_project_is_trained,
    ensure_trained_anomaly_project,
    ensure_trained_example_project,
)
from .video_helpers import AsyncVideoProcessor, predict_video_from_deployment

__all__ = [
    "DEFAULT_DATA_PATH",
    "EXAMPLE_IMAGE_PATH",
    "create_segmentation_demo_project",
    "create_anomaly_classification_demo_project",
    "create_classification_demo_project",
    "create_detection_demo_project",
    "create_detection_to_segmentation_demo_project",
    "create_detection_to_classification_demo_project",
    "ensure_trained_example_project",
    "ensure_trained_anomaly_project",
    "ensure_project_is_trained",
    "get_coco_dataset",
    "get_mvtec_dataset",
    "set_directory_permissions",
    "get_person_car_bike_video",
    "predict_video_from_deployment",
    "AsyncVideoProcessor",
]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
from datetime import datetime
from typing import List, Optional, Union

from geti_sdk.data_models import CreditAccount, CreditBalance, Subscription
from geti_sdk.data_models.job import Job, JobCost
from geti_sdk.http_session import GetiSession
from geti_sdk.utils.job_helpers import get_job_by_id
from geti_sdk.utils.serialization_helpers import deserialize_dictionary
from geti_sdk.utils.workspace_helpers import get_default_workspace_id


def allow_supported(func):
    """
    Decorate the class methods to allow them to run only if the Credit System is supported.
    """

    def wrapper(instance, *args, **kwargs):
        if instance._is_supported:
            return func(instance, *args, **kwargs)
        else:
            logging.warning(
                "Credit System is not supported by the Intel Geti Platform."
            )
            return None

    return wrapper


class CreditSystemClient:
    """
    Class to work with credits in Intel Geti.
    """

    def __init__(self, session: GetiSession, workspace_id: Optional[str] = None):
        self.session = session
        if workspace_id is not None:
            self.workspace_id = workspace_id
        else:
            self.workspace_id = get_default_workspace_id(self.session)
        # Make sure the Intel Geti Platform supports Credit System.
        self._is_supported = self.is_supported()

    def is_supported(self) -> bool:
        """
        Check if the Intel Geti Platform supports Credit system.

        :return: True if the Credit System is supported, False otherwise.
        """
        # Send a GET request to the balance endpoint to check if the Credit System is supported.
        # The text response is allowed to check if the server responds with a default
        # html page in case the Credit System is not supported.
        r = self.session.get_rest_response(
            url=self.session.base_url + "balance",
            method="GET",
            allow_text_response=True,
        )
        if isinstance(r, dict):
            # If the Platform responds with the information about the available subscriptions,
            # then it supports Credit System. Session will decode the message into a `dict`
            return True
        else:
            # In case the server returns an empty 200 message, it means the Credit System is not supported.
            # The session would return a Response object.
            return False

    @allow_supported
    def get_balance(
        self, timestamp: Optional[datetime] = None
    ) -> Optional[CreditBalance]:
        """
        Get the current credit balance in the workspace.

        :param timestamp: The timestamp to get the balance at. If None, the current balance is returned.
        :return: The available credit balance in the workspace.
        """
        query_postfix = (
            f"?date={int(timestamp.timestamp() * 1000)}" if timestamp else ""
        )
        response = self.session.get_rest_response(
            url=self.session.base_url + "balance" + query_postfix,
            method="GET",
        )
        return deserialize_dictionary(response, CreditBalance)

    def get_job_cost(self, job: Union[Job, str]) -> Optional[JobCost]:
        """
        Get the cost of a job.

        This method allows you to find out the cost of a training or an optimization job.

        :param job: A Job object or a Job ID.
        :return: A JobCost object presenting the total cost and the consumed credits.
        """
        if isinstance(job, Job):
            job_id = job.id
        else:
            job_id = job
        fetched_job = get_job_by_id(job_id, self.session, self.workspace_id)
        return fetched_job.cost if fetched_job else None

    @allow_supported
    def get_subscriptions(self) -> Optional[List[Subscription]]:
        """
        Get the subscription details for the workspace.

        :return: A list of Subscription objects.
        """
        response = self.session.get_rest_response(
            url=self.session.base_url + "subscriptions",
            method="GET",
        )
        return [
            deserialize_dictionary(sub, Subscription)
            for sub in response.get("subscriptions", [])
        ]

    @allow_supported
    def get_credit_accounts(self) -> Optional[List[CreditAccount]]:
        """
        Get the subscription details for the workspace.

        :return: A list of Subscription objects.
        """
        response = self.session.get_rest_response(
            url=self.session.base_url + "credit_accounts",
            method="GET",
        )
        return [
            deserialize_dictionary(sub, CreditAccount)
            for sub in response.get("credit_accounts", [])
        ]
